@@ -9,13 +9,99 @@ would appreciate credit if this program or parts of it are used.
 */
 
 #include <stdio.h>
+
+#include "Dbg_cf.h"
+#if 0
+/* tclInt.h drags in stdlib.  By claiming no-stdlib, force it to drag in */
+/* Tcl's compat version.  This avoids having to test for its presence */
+/* which is too tricky - configure can't generate two cf files, so when */
+/* Expect (or any app) uses the debugger, there's no way to get the info */
+/* about whether stdlib exists or not, except pointing the debugger at */
+/* an app-dependent .h file and I don't want to do that. */
+#define NO_STDLIB_H
+#endif
+
+
 #include "tclInt.h"
 /*#include <varargs.h>		tclInt.h drags in varargs.h.  Since Pyramid */
 /*				objects to including varargs.h twice, just */
 /*				omit this one. */
-#include "string.h"
+/*#include "string.h"		tclInt.h drags this in, too! */
 #include "Dbg.h"
 
+
+/* START OF SECTION FROM expect_comm.h.  */
+
+/*
+ * Definitions that allow Tcl functions with variable numbers of
+ * arguments to be used with either varargs.h or stdarg.h.  TCL_VARARGS
+ * is used in procedure prototypes.  TCL_VARARGS_DEF is used to declare
+ * the arguments in a function definiton: it takes the type and name of
+ * the first argument and supplies the appropriate argument declaration
+ * string for use in the function definition.  TCL_VARARGS_START
+ * initializes the va_list data structure and returns the first argument.  */
+
+/* in Tcl 7.5, Tcl now supplies these definitions */
+#if defined(TCL_VARARGS)
+#   define EXP_VARARGS		TCL_VARARGS
+#   define EXP_VARARGS_DEF	TCL_VARARGS_DEF
+#   define EXP_VARARGS_START	TCL_VARARGS_START
+#else
+	/* Tcl's test for this is simply wrong */
+#  if defined(HAVE_STDARG_H)
+#   define EXP_VARARGS(type, name) (type name, ...)
+#   define EXP_VARARGS_DEF(type, name) (type name, ...)
+#   define EXP_VARARGS_START(type, name, list) (va_start(list, name), name)
+#  else
+#   define EXP_VARARGS(type, name) ()
+#   define EXP_VARARGS_DEF(type, name) (va_alist) va_dcl
+#   define EXP_VARARGS_START(type, name, list) \
+	(va_start(list), va_arg(list, type))
+#  endif /* defined(HAVE_STDARG_H) */
+#endif /* defined(TCL_VARARGS) */
+
+/*
+ * Definitions that allow this header file to be used either with or
+ * without ANSI C features like function prototypes.
+ */
+
+#if ((defined(__STDC__) || defined(SABER)) && !defined(NO_PROTOTYPE)) || defined(__cplusplus) || defined(USE_PROTOTYPE)
+#   define _USING_PROTOTYPES_ 1
+#   define EXP_PROTO(x)	x
+#else
+#   define EXP_PROTO(x)	()
+#endif
+
+/* support for Standard C and C++ prototypes */
+#ifdef __cplusplus
+#  define EXP_PROTOTYPES
+#  if !defined(EXP_VARARGS)
+#  define EXP_VARARGS	...
+#  endif
+#else
+#  if !defined(EXP_VARARGS)
+#  define EXP_VARARGS	, ...
+#  endif
+#  ifdef __STDC__
+#    define EXP_PROTOTYPES
+#  endif /*__STDC__*/
+#endif /*__cplusplus*/
+
+#ifdef EXP_PROTOTYPES
+#define EXP_PROTO(x)	x
+#ifdef EXP_DEFINE_FNS
+/* when functions are really being defined, we have to use va_alist as arg */
+#define EXP_PROTOV(x)	va_alist
+#else
+#define EXP_PROTOV(x)	x
+#endif /*EXP_DEFINE_FNS*/
+#else
+#define EXP_PROTO(x)	()
+#define EXP_PROTOV(x)	()
+#endif /*EXP_PROTOTYPES*/
+
+/* END OF SECTION FROM expect_comm.h.  */
+
 #ifndef TRUE
 #define TRUE 1
 #define FALSE 0
@@ -28,10 +114,12 @@ static int zero();
 /* moved into Tcl_Interp */
 
 static Dbg_InterProc *interactor = simple_interactor;
+static ClientData interdata = 0;
 static Dbg_IgnoreFuncsProc *ignoreproc = zero;
 static Dbg_OutputProc *printproc = 0;
+static ClientData printdata = 0;
 
-static void print();
+static void print EXP_PROTOV(EXP_VARARGS(Tcl_Interp*,arg1));
 
 static int debugger_active = FALSE;
 
@@ -128,7 +216,7 @@ struct breakpoint *b;
 	if (b->cmd)
 		print(interp,"then {%s}",b->cmd);
 
-	putchar('\n');
+	print(interp,"\n");
 }
 
 static void
@@ -181,7 +269,8 @@ struct breakpoint *bp;	/* breakpoint to test */
 
 		/* ignore errors, since they are likely due to */
 		/* simply being out of scope a lot */
-		if (TCL_OK != Tcl_ExprBoolean(interp,bp->expr,&value)) return 0;
+		if (TCL_OK != Tcl_ExprBoolean(interp,bp->expr,&value)
+		    || (value == 0)) return 0;
 	}
 
 	if (bp->cmd) {
@@ -525,20 +614,24 @@ char *argv[];
 	/* save so we can restore later */
 	trueFramePtr = iPtr->varFramePtr;
 
-	/* test all breakpoints to see if we should break */
+	/* do not allow breaking while testing breakpoints */
 	debug_suspended = TRUE;
 
+	/* test all breakpoints to see if we should break */
 	/* if any successful breakpoints, start interactor */
 	debug_new_action = FALSE;	/* reset strobe */
 	break_status = FALSE;		/* no successful breakpoints yet */
 	for (b = break_base;b;b=b->next) {
 		break_status |= breakpoint_test(interp,command,b);
 	}
-	if (!debug_new_action && break_status) goto start_interp;
+	if (break_status) {
+		if (!debug_new_action) goto start_interact;
 
-	/* if s or n triggered by breakpoint, make "s 1" (and so on) */
-	/* refer to next command, not this one */	
-	if (debug_new_action) step_count++;
+		/* if s or n triggered by breakpoint, make "s 1" */
+		/* (and so on) refer to next command, not this one */
+/*		step_count++;*/
+		goto end_interact;
+	}
 
 	switch (debug_cmd) {
 	case cont:
@@ -546,7 +639,7 @@ char *argv[];
 	case step:
 		step_count--;
 		if (step_count > 0) goto finish;
-		goto start_interp;
+		goto start_interact;
 	case next:
 		/* check if we are back at the same level where the next */
 		/* command was issued.  Also test */
@@ -556,21 +649,21 @@ char *argv[];
 		if (GoalFrame(goalFramePtr,iPtr)) goto finish;
 		step_count--;
 		if (step_count > 0) goto finish;
-		goto start_interp;
+		goto start_interact;
 	case Next:
 		/* check if we are back at the same level where the next */
 		/* command was issued.  */
 		if (goalNumLevel < iPtr->numLevels) goto finish;
 		step_count--;
 		if (step_count > 0) goto finish;
-		goto start_interp;
+		goto start_interact;
 	case ret:
 		/* same comment as in "case next" */
 		if (goalFramePtr != iPtr->varFramePtr) goto finish;
-		goto start_interp;
+		goto start_interact;
 	}
 
-start_interp:
+start_interact:
 	if (print_command_first_time) {
 		print(interp,"%s: %s\n",
 				level_text,print_argv(interp,1,&command));
@@ -578,10 +671,11 @@ start_interp:
 	}
 	/* since user is typing a command, don't interrupt it immediately */
 	debug_cmd = cont;
-	debug_suspended = FALSE;
+	debug_suspended = TRUE;
 
 	/* interactor won't return until user gives a debugger cmd */
-	(*interactor)(interp);
+	(*interactor)(interp,interdata);
+end_interact:
 
 	/* save this so it can be restored after "w" command */
 	viewFramePtr = iPtr->varFramePtr;
@@ -593,14 +687,16 @@ start_interp:
 			print(interp,"%s\n",interp->result);
 			Tcl_ResetResult(interp);
 		}
-		goto start_interp;
+		goto start_interact;
 	}
 
 	/* reset view back to normal */
 	iPtr->varFramePtr = trueFramePtr;
 
+#if 0
 	/* allow trapping */
 	debug_suspended = FALSE;
+#endif
 
 	switch (debug_cmd) {
 	case cont:
@@ -627,7 +723,7 @@ start_interp:
 
 	/* restore view and restart interactor */
 	iPtr->varFramePtr = viewFramePtr;
-	goto start_interp;
+	goto start_interact;
 
  finish:
 	debug_suspended = FALSE;
@@ -951,15 +1047,17 @@ char **argv;
 #define PAD 80
 
 static void
-print(va_alist)
-va_dcl
+print EXP_VARARGS_DEF(Tcl_Interp*,arg1)
+/*print(va_alist)*/
+/*va_dcl*/
 {
 	Tcl_Interp *interp;
 	char *fmt;
 	va_list args;
 
-	va_start(args);
-	interp = va_arg(args,Tcl_Interp *);
+	interp = EXP_VARARGS_START(Tcl_Interp*,arg1,args);
+	/*va_start(args);*/
+	/*interp = va_arg(args,Tcl_Interp *);*/
 	fmt = va_arg(args,char *);
 	if (!printproc) vprintf(fmt,args);
 	else {
@@ -974,19 +1072,24 @@ va_dcl
 		}
 
 		vsprintf(buf,fmt,args);
-		(*printproc)(interp,buf);
+		(*printproc)(interp,buf,printdata);
 	}
 	va_end(args);
 }
 
 /*ARGSUSED*/
-Dbg_InterProc *
-Dbg_Interactor(interp,inter_proc)
+Dbg_InterStruct
+Dbg_Interactor(interp,inter_proc,data)
 Tcl_Interp *interp;
 Dbg_InterProc *inter_proc;
+ClientData data;
 {
-	Dbg_InterProc *tmp = interactor;
+	Dbg_InterStruct tmp;
+
+	tmp.func = interactor;
+	tmp.data = interdata;
 	interactor = (inter_proc?inter_proc:simple_interactor);
+	interdata = data;
 	return tmp;
 }
 
@@ -1002,17 +1105,21 @@ Dbg_IgnoreFuncsProc *proc;
 }
 
 /*ARGSUSED*/
-Dbg_OutputProc *
-Dbg_Output(interp,proc)
+Dbg_OutputStruct
+Dbg_Output(interp,proc,data)
 Tcl_Interp *interp;
 Dbg_OutputProc *proc;
+ClientData data;
 {
-	Dbg_OutputProc *tmp = printproc;
-	printproc = (proc?proc:0);
+	Dbg_OutputStruct tmp;
+
+	tmp.func = printproc;
+	tmp.data = printdata;
+	printproc = proc;
+	printdata = data;
 	return tmp;
 }
 
-
 /*ARGSUSED*/
 int
 Dbg_Active(interp)
@@ -1143,7 +1250,12 @@ Tcl_Interp *interp;
 			}
 		}
 
+#if TCL_MAJOR_VERSION == 7 && TCL_MINOR_VERSION < 4
 		rc = Tcl_RecordAndEval(interp,ccmd,0);
+#else
+		rc = Tcl_RecordAndEval(interp,ccmd,TCL_NO_EVAL);
+		rc = Tcl_Eval(interp,ccmd);
+#endif
 #if TCL_MAJOR_VERSION != 6
 		Tcl_DStringFree(&dstring);
 #endif
@@ -1181,6 +1293,8 @@ Tcl_Interp *interp;
 	return(rc);
 }
 
+static char init_auto_path[] = "lappend auto_path $dbg_library";
+
 static void
 init_debugger(interp)
 Tcl_Interp *interp;
@@ -1197,6 +1311,11 @@ Tcl_Interp *interp;
 
 	debugger_active = TRUE;
 	Tcl_SetVar2(interp,Dbg_VarName,"active","1",0);
+#ifdef DBG_SCRIPTDIR
+	Tcl_SetVar(interp,"dbg_library",DBG_SCRIPTDIR,0);
+#endif
+	Tcl_Eval(interp,init_auto_path);
+
 }
 
 /* allows any other part of the application to jump to the debugger */
