@@ -12,15 +12,44 @@ would appreciate credit if this program or parts of it are used.
 #include <setjmp.h>
 #include <sys/types.h>
 #include <sys/ioctl.h>
+
+#ifdef CRAY
+# ifndef TCSETCTTY
+#  if defined(HAVE_TERMIOS)
+#   include <termios.h>
+#  else
+#   include <termio.h>
+#  endif
+# endif
+#endif
+
 #ifdef HAVE_SYS_FCNTL_H
 #  include <sys/fcntl.h>
 #else
 #  include <fcntl.h>
 #endif
+
+#ifdef HAVE_STRREDIR_H
+#include <sys/strredir.h>
+# ifdef SRIOCSREDIR
+#  undef TIOCCONS
+# endif
+#endif
+
 #include <signal.h>
 /*#include <memory.h> - deprecated - ANSI C moves them into string.h */
 #include "string.h"
-#include <varargs.h>
+
+#if 0 /* This is now handled in expect_comm.h.  */
+#ifdef HAVE_VARARGS_H
+#  include <varargs.h>
+#else
+#  ifdef HAVE_STDARG_H
+#    include <stdarg.h>
+#  endif
+#endif
+#endif /* 0 */
+
 #include <errno.h>
 #include "exp_rename.h"
 #define EXP_DEFINE_FNS
@@ -29,12 +58,15 @@ would appreciate credit if this program or parts of it are used.
 
 #include "exp_printify.h"
 
-/*#ifndef _STDLIB*/
-/*char *malloc();*/
-/*void exit();*/
-/*#endif*/
-#include "stdlib.h"
-/*extern char *realloc();*/
+#ifdef NO_STDLIB_H
+#  if TCL_MAJOR_VERSION == 7 && TCL_MINOR_VERSION >= 5
+#    include "../compat/stdlib.h"
+#  else
+#    include "compat/stdlib.h"
+#  endif
+#else
+#  include <stdlib.h>		/* for malloc */
+#endif /*NO_STDLIB_H*/
 
 #define EXP_MATCH_MAX	2000
 /* public */
@@ -54,6 +86,7 @@ char *exp_stty_init = 0;		/* initial stty args */
 int exp_ttycopy = TRUE;			/* copy tty parms from /dev/tty */
 int exp_ttyinit = TRUE;			/* set tty parms to sane state */
 int exp_console = FALSE;		/* redirect console */
+void (*exp_child_exec_prelude)() = 0;
 
 jmp_buf exp_readenv;		/* for interruptable read() */
 int exp_reading = FALSE;	/* whether we can longjmp or not */
@@ -67,7 +100,75 @@ int Exp_StringMatch();
 
 void exp_init_pty();
 
+#ifndef HAVE_TCLREGERROR
 extern char *tclRegexpError;	/* declared in tclInt.h */
+#else
+extern char *TclGetRegError();
+extern void TclRegError();
+#endif
+
+static unsigned int bufsiz = 2*EXP_MATCH_MAX;
+
+static struct f {
+	int valid;
+
+	char *buffer;		/* buffer of matchable chars */
+	char *buffer_end;	/* one beyond end of matchable chars */
+	/*char *match;		/* start of matched string */
+	char *match_end;	/* one beyond end of matched string */
+	int msize;		/* size of allocate space */
+				/* actual size is one larger for null */
+} *fs = 0;
+
+static int fd_alloc_max = -1;	/* max fd allocated */
+
+/* translate fd or fp to fd */
+static struct f *
+fdfp2f(fd,fp)
+int fd;
+FILE *fp;
+{
+	if (fd == -1) return(fs + fileno(fp));
+	else return(fs + fd);
+}
+
+static struct f *
+fd_new(fd)
+int fd;
+{
+	int i, low;
+	struct f *fp;
+	struct f *newfs;	/* temporary, so we don't lose old fs */
+
+	if (fd > fd_alloc_max) {
+		if (!fs) {	/* no fd's yet allocated */
+			newfs = (struct f *)malloc(sizeof(struct f)*(fd+1));
+			low = 0;
+		} else {		/* enlarge fd table */
+			newfs = (struct f *)realloc((char *)fs,sizeof(struct f)*(fd+1));
+			low = fd_alloc_max+1;
+		}
+		fs = newfs;
+		fd_alloc_max = fd;
+		for (i = low; i <= fd_alloc_max; i++) { /* init new entries */
+			fs[i].valid = FALSE;
+		}
+	}
+
+	fp = fs+fd;
+
+	if (!fp->valid) {
+		/* initialize */
+		fp->buffer = malloc((unsigned)(bufsiz+1));
+		if (!fp->buffer) return 0;
+		fp->msize = bufsiz;
+		fp->valid = TRUE;
+	}
+	fp->buffer_end = fp->buffer;
+	fp->match_end = fp->buffer;
+	return fp;
+
+}
 
 /* returns fd of master side of pty */
 int
@@ -76,8 +177,11 @@ char *file;
 char *argv[];	/* some compiler complains about **argv? */
 {
 	int cc;
+	int errorfd;	/* place to stash fileno(stderr) in child */
+			/* while we're setting up new stderr */
 	int ttyfd;
 	int sync_fds[2];
+	int sync2_fds[2];
 	char sync_byte;
 #ifdef PTYTRAP_DIES
 	int slave_write_ioctls = 1;
@@ -89,6 +193,7 @@ char *argv[];	/* some compiler complains about **argv? */
 	if (first_time) {
 		first_time = FALSE;
 		exp_init_pty();
+		exp_init_tty();
 	}
 
 	if (!file || !argv) sysreturn(EINVAL);
@@ -114,12 +219,23 @@ when trapping, see below in child half of fork */
 	exp_slave_control(exp_pty[0],1);*/
 #endif
 
-	pipe(sync_fds);
+	if (!fd_new(exp_pty[0])) {
+		errno = ENOMEM;
+		return -1;
+	}
+
+	if (-1 == (pipe(sync_fds))) {
+		return -1;
+	}
+	if (-1 == (pipe(sync2_fds))) {
+		return -1;
+	}
 
 	if ((exp_pid = fork()) == -1) return(-1);
 	if (exp_pid) {
 		/* parent */
 		close(sync_fds[1]);
+		close(sync2_fds[0]);
 		if (!exp_autoallocpty) close(exp_pty[1]);
 
 #ifdef PTYTRAP_DIES
@@ -145,7 +261,7 @@ when trapping, see below in child half of fork */
 			}
 		}
 #endif
-#endif PTYTRAP_DIES
+#endif /*PTYTRAP_DIES*/
 
 		/*
 		 * wait for slave to initialize pty before allowing
@@ -160,13 +276,23 @@ when trapping, see below in child half of fork */
 		}
 
 		/* turn on detection of eof */
-		/* this should be done as quickly as possible.  Conceivably */
-		/* child could die quickly enough that we are not ready */
-		/* for the eof.  Fortunately, the child has a lot of work */
-		/* to do so this seems unlikely. */
 		exp_slave_control(exp_pty[0],1);
 
+		/*
+		 * tell slave to go on now now that we have initialized pty
+		 */
+
+		debuglog("parent: telling child to go ahead\r\n");
+		cc = write(sync2_fds[1]," ",1);
+		if (cc == -1) {
+			errorlog("parent sync byte write: %s\r\n",strerror(errno));
+			exit(-1);
+		}
+
+		debuglog("parent: now unsynchronized from child\r\n");
+
 		close(sync_fds[0]);
+		close(sync2_fds[1]);
 
 		return(exp_pty[0]);
 	}
@@ -174,9 +300,10 @@ when trapping, see below in child half of fork */
 	/* child process - do not return from here!  all errors must exit() */
 
 	close(sync_fds[0]);
+	close(sync2_fds[1]);
 
 #ifdef CRAY
-	(void) close(master);
+	(void) close(exp_pty[0]);
 #endif
 
 /* ultrix (at least 4.1-2) fails to obtain controlling tty if setsid */
@@ -215,32 +342,38 @@ when trapping, see below in child half of fork */
 #endif /* SYSV3 */
 #endif /* DO_SETSID */
 
+	/* save error fd while we're setting up new one */
+	errorfd = fcntl(2,F_DUPFD,3);
+	/* and here is the macro to restore it */
+#define restore_error_fd {close(2);fcntl(errorfd,F_DUPFD,2);}
 
 	if (exp_autoallocpty) {
+
 	    close(0);
 	    close(1);
-	    /* leave 2 around awhile for stderr-related stuff */
+	    close(2);
 
 	    /* since we closed fd 0, open of pty slave must return fd 0 */
 
 	    if (0 > (exp_pty[1] = getptyslave(exp_ttycopy,exp_ttyinit,
 						exp_stty_init))) {
+		restore_error_fd
 		fprintf(stderr,"open(slave pty): %s\n",strerror(errno));
 		exit(-1);
 	    }
 	    /* sanity check */
 	    if (exp_pty[1] != 0) {
+		restore_error_fd
 		fprintf(stderr,"getptyslave: slave = %d but expected 0\n",
 								exp_pty[1]);
 		exit(-1);
 	    }
 	} else {
 		if (exp_pty[1] != 0) {
-			close(0);
-			fcntl(exp_pty[1],F_DUPFD,0);
+			close(0);	fcntl(exp_pty[1],F_DUPFD,0);
 		}
-		close(1);
-		fcntl(0,F_DUPFD,1);
+		close(1);		fcntl(0,F_DUPFD,1);
+		close(2);		fcntl(0,F_DUPFD,1);
 		close(exp_pty[1]);
 	}
 
@@ -250,10 +383,11 @@ when trapping, see below in child half of fork */
 /* code should be skipped on the hp9000s300 and hp9000s720 (but there */
 /* is no documented define for the 720!) */
 
-#if defined(TIOCSCTTY) && !defined(CIBAUD) && !defined(sun) && !defined(hpux)
+#if defined(TIOCSCTTY) && !defined(sun) && !defined(hpux)
 	/* 4.3+BSD way to acquire controlling terminal */
 	/* according to Stevens - Adv. Prog..., p 642 */
 	if (ioctl(0,TIOCSCTTY,(char *)0) < 0) {
+		restore_error_fd
 		fprintf(stderr,"failed to get controlling terminal using TIOCSCTTY");
 		exit(-1);
 	}
@@ -264,6 +398,7 @@ when trapping, see below in child half of fork */
  	(void) ioctl(0,TCSETCTTY,0);
  	(void) close(0);
  	if (open("/dev/tty", O_RDWR) < 0) {
+		restore_error_fd
  		fprintf(stderr,"open(/dev/tty): %s\r\n",strerror(errno));
  		exit(-1);
  	}
@@ -283,6 +418,7 @@ when trapping, see below in child half of fork */
 	 * least).
 	 */
 	if ((pid = fork()) == -1) {
+		restore_error_fd
 		fprintf(stderr,"second fork: %s\r\n",strerror(errno));
 		exit(-1);
 	}
@@ -313,19 +449,31 @@ when trapping, see below in child half of fork */
 #endif /* _CRAY2 */
 #endif /* CRAY */
 
-#ifdef TIOCCONS
 	if (exp_console) {
-		int on = 1;
-		if (ioctl(0,TIOCCONS,(char *)&on) == -1) {
+#ifdef SRIOCSREDIR
+		int fd;
+
+		if ((fd = open("/dev/console", O_RDONLY)) == -1) {
+			restore_error_fd
 			fprintf(stderr, "spawn %s: cannot open console, check permissions of /dev/console\n",argv[0]);
 			exit(-1);
 		}
-	}
-#endif /* TIOCCONS */
+		if (ioctl(fd, SRIOCSREDIR, 0) == -1) {
+			restore_error_fd
+			fprintf(stderr, "spawn %s: cannot redirect console, check permissions of /dev/console\n",argv[0]);
+		}
+		close(fd);
+#endif
 
-	/* by now, fd 0 and 1 point to slave pty, so fix 2 */
-	close(2);
-	fcntl(0,F_DUPFD,2);	/* duplicate 0 onto 2 */
+#ifdef TIOCCONS
+		int on = 1;
+		if (ioctl(0,TIOCCONS,(char *)&on) == -1) {
+			restore_error_fd
+			fprintf(stderr, "spawn %s: cannot open console, check permissions of /dev/console\n",argv[0]);
+			exit(-1);
+		}
+#endif /* TIOCCONS */
+	}
 
 	/* tell parent that we are done setting up pty */
 	/* The actual char sent back is irrelevant. */
@@ -333,17 +481,35 @@ when trapping, see below in child half of fork */
 	/* debuglog("child: telling parent that pty is initialized\r\n");*/
 	cc = write(sync_fds[1]," ",1);
 	if (cc == -1) {
+		restore_error_fd
 		fprintf(stderr,"child: sync byte write: %s\r\n",strerror(errno));
 		exit(-1);
 	}
 	close(sync_fds[1]);
 
+	/* wait for master to let us go on */
+	/* debuglog("child: waiting for go ahead from parent\r\n"); */
+
+/*	close(master);	/* force master-side close so we can read */
+	cc = read(sync2_fds[0],&sync_byte,1);
+	if (cc == -1) {
+		restore_error_fd
+		errorlog("child: sync byte read: %s\r\n",strerror(errno));
+		exit(-1);
+	}
+	close(sync2_fds[0]);
+
+	/* debuglog("child: now unsynchronized from parent\r\n"); */
+
 	/* (possibly multiple) masters are closed automatically due to */
 	/* earlier fcntl(,,CLOSE_ON_EXEC); */
 
 	/* just in case, allow user to explicitly close other files */
 	if (exp_close_in_child) (*exp_close_in_child)();
 
+	/* allow user to do anything else to child */
+	if (exp_child_exec_prelude) (*exp_child_exec_prelude)();
+
         (void) execvp(argv[0],argv);
 	/* Unfortunately, by now we've closed fd's to stderr, logfile and
 		debugfile.
@@ -359,23 +525,26 @@ when trapping, see below in child half of fork */
 /* returns fd of master side of pty */
 /*VARARGS*/
 int
-exp_spawnl(va_alist)
-va_dcl
+exp_spawnl EXP_VARARGS_DEF(char *,arg1)
+/*exp_spawnl(va_alist)*/
+/*va_dcl*/
 {
 	va_list args;
 	int i;
 	char *arg, **argv;
 
-	va_start(args);
-	for (i=0;;i++) {
+	arg = EXP_VARARGS_START(char *,arg1,args);
+	/*va_start(args);*/
+	for (i=1;;i++) {
 		arg = va_arg(args,char *);
 		if (!arg) break;
 	}
 	va_end(args);
 	if (i == 0) sysreturn(EINVAL);
 	if (!(argv = (char **)malloc((i+1)*sizeof(char *)))) sysreturn(ENOMEM);
-	va_start(args);
-	for (i=0;;i++) {
+	argv[0] = EXP_VARARGS_START(char *,arg1,args);
+	/*va_start(args);*/
+	for (i=1;;i++) {
 		argv[i] = va_arg(args,char *);
 		if (!argv[i]) break;
 	}
@@ -384,6 +553,18 @@ va_dcl
 	return(i);
 }
 
+/* allow user-provided fd to be passed to expect funcs */
+int
+exp_spawnfd(fd)
+int fd;
+{
+	if (!fd_new(fd)) {
+		errno = ENOMEM;
+		return -1;
+	}
+	return fd;	
+}
+
 /* remove nulls from s.  Initially, the number of chars in s is c, */
 /* not strlen(s).  This count does not include the trailing null. */
 /* returns number of nulls removed. */
@@ -446,8 +627,7 @@ int timeout;
 	/* abort if setjmp returns 1 (EXP_ABORT). */
 	if (EXP_ABORT != setjmp(exp_readenv)) {
 		exp_reading = TRUE;
-		if (fd != -1) cc = read(fd,buffer,length);
-		else {
+		if (fd == -1) {
 			int c;
 			c = getc(fp);
 			if (c == EOF) {
@@ -457,6 +637,42 @@ int timeout;
 				buffer[0] = c;
 				cc = 1;
 			}
+		} else {
+#ifndef HAVE_PTYTRAP
+			cc = read(fd,buffer,length);
+#else
+#  include <sys/ptyio.h>
+
+			fd_set rdrs;
+			fd_set excep;
+
+		restart:
+			FD_ZERO(&rdrs);
+			FD_ZERO(&excep);
+			FD_SET(fd,&rdrs);
+			FD_SET(fd,&excep);
+			if (-1 == (cc = select(fd+1,
+					 (SELECT_MASK_TYPE *)&rdrs,
+					 (SELECT_MASK_TYPE *)0,
+					 (SELECT_MASK_TYPE *)&excep,
+					 (struct timeval *)0))) {
+				/* window refreshes trigger EINTR, ignore */
+				if (errno == EINTR) goto restart;
+			}
+			if (FD_ISSET(fd,&rdrs)) {
+				cc = read(fd,buffer,length);
+			} else if (FD_ISSET(fd,&excep)) {
+				struct request_info ioctl_info;
+				ioctl(fd,TIOCREQCHECK,&ioctl_info);
+				if (ioctl_info.request == TIOCCLOSE) {
+					cc = 0; /* indicate eof */
+				} else {
+					ioctl(fd, TIOCREQSET, &ioctl_info);
+					/* presumably, we trapped an open here */
+					goto restart;
+				}
+			}
+#endif /* HAVE_PTYTRAP */
 		}
 #if 0
 		/* can't get fread to return early! */
@@ -475,8 +691,6 @@ int timeout;
 	return(cc);
 }
 
-static unsigned int bufsiz = 2*EXP_MATCH_MAX;
-
 /* I tried really hard to make the following two functions share the code */
 /* that makes the ecase array, but I kept running into a brick wall when */
 /* passing var args into the funcs and then again into a make_cases func */
@@ -501,42 +715,57 @@ struct exp_case *ecases;
 	int cc = 0;		/* number of chars returned in a single read */
 	int buf_length;		/* numbers of chars in exp_buffer */
 	int old_length;		/* old buf_length */
+	int first_time = TRUE;	/* force old buffer to be tested before */
+				/* additional reads */
+	int polled = 0;		/* true if poll has caused read() to occur */
 
-	unsigned int new_siz;	/* this is just match_max*2 for efficiency */
-	char *new_buf;
 	struct exp_case *ec;	/* points to current ecase */
 
 	time_t current_time;	/* current time (when we last looked)*/
 	time_t end_time;	/* future time at which to give up */
 	int remtime;		/* remaining time in timeout */
 
-	if (exp_buffer == 0) {
-		if (!(exp_buffer = malloc((unsigned)(bufsiz+1))))
-			sysreturn(ENOMEM);
-		exp_buffer_end = exp_buffer;
-		exp_match_end = exp_buffer;
-	}
+	struct f *f;
+	int return_val;
+	int sys_error = 0;
+#define return_normally(x)	{return_val = x; goto cleanup;}
+#define return_errno(x)	{sys_error = x; goto cleanup;}
+
+	f = fdfp2f(fd,fp);
+	if (!f) return_errno(ENOMEM);
+
+	exp_buffer = f->buffer;
+	exp_buffer_end = f->buffer_end;
+	exp_match_end = f->match_end;
 
 	buf_length = exp_buffer_end - exp_match_end;
 	if (buf_length) {
 		/*
-		 *take end of previous match to end of buffer
+		 * take end of previous match to end of buffer
 		 * and copy to beginning of buffer
 		 */
-		memcpy(exp_buffer,exp_match_end,buf_length);
+		memmove(exp_buffer,exp_match_end,buf_length);
 	}			
 	exp_buffer_end = exp_buffer + buf_length;
 	*exp_buffer_end = '\0';
 
-	if (!ecases) sysreturn(EINVAL);
+	if (!ecases) return_errno(EINVAL);
 
 	/* compile if necessary */
 	for (ec=ecases;ec->type != exp_end;ec++) {
 		if ((ec->type == exp_regexp) && !ec->re) {
+#ifndef HAVE_TCLREGERROR
 			tclRegexpError = 0;
+#else
+			TclRegError((char *)0);
+#endif
 			if (!(ec->re = TclRegComp(ec->pattern))) {
+#ifndef HAVE_TCLREGERROR
 				fprintf(stderr,"regular expression %s is bad: %s",ec->pattern,tclRegexpError);
-				sysreturn(EINVAL);
+#else
+				fprintf(stderr,"regular expression %s is bad: %s",ec->pattern,TclGetRegError());
+#endif
+				return_errno(EINVAL);
 			  }
 		  }
 	}
@@ -545,12 +774,26 @@ struct exp_case *ecases;
 	/* reasons.  1) Need twice the space in case the match */
 	/* straddles two bufferfuls, 2) easier to hack the division by */
 	/* two when shifting the buffers later on */
-	if (bufsiz != (new_siz = 2*exp_match_max)) {
-		if (0 == (new_buf = realloc(exp_buffer,new_siz+1)))
-			sysreturn(ENOMEM);
-		bufsiz = new_siz;
-		exp_buffer = new_buf;
+
+	bufsiz = 2*exp_match_max;
+	if (f->msize != bufsiz) {
+		/* if truncated, forget about some data */
+		if (buf_length > bufsiz) {
+			/* copy end of buffer down */
+
+			/* copy one less than what buffer can hold to avoid */
+			/* triggering buffer-full handling code below */
+			/* which will immediately dump the first half */
+			/* of the buffer */
+			memmove(exp_buffer,exp_buffer+(buf_length - bufsiz)+1,
+				bufsiz-1);
+			buf_length = bufsiz-1;
+		}
+		exp_buffer = ckrealloc(exp_buffer,bufsiz+1);
+		if (!exp_buffer) return_errno(ENOMEM);
+		exp_buffer[buf_length] = '\0';
 		exp_buffer_end = exp_buffer + buf_length;
+		f->msize = bufsiz;
 	}
 
 	if (exp_timeout != -1) signal(SIGALRM,sigalarm_handler);
@@ -559,9 +802,6 @@ struct exp_case *ecases;
 	remtime = exp_timeout;
 
 	time(&current_time);
-	/* if user sets timeout to 0 (i.e. poll), this indicates */
-	/* we should do one read and return */
-	/* Hopefully, the buffer is big enough and will get it all! */
 	end_time = current_time + remtime;
 
 	for (;;) {
@@ -570,8 +810,13 @@ struct exp_case *ecases;
 		if (buf_length == bufsiz) {
 			int first_half, second_half;
 
-			if (exp_full_buffer) return EXP_FULLBUFFER;
-
+			if (exp_full_buffer) {
+				debuglog("expect: full buffer\r\n");
+				exp_match = exp_buffer;
+				exp_match_end = exp_buffer + buf_length;
+				exp_buffer_end = exp_match_end;
+				return_normally(EXP_FULLBUFFER);
+			}
 			first_half = bufsiz/2;
 			second_half = bufsiz - first_half;
 
@@ -580,24 +825,53 @@ struct exp_case *ecases;
 			exp_buffer_end = exp_buffer + second_half;
 		}
 
-		if ((exp_timeout != -1) && (remtime < 0)) {
-			return EXP_TIMEOUT;
-		} else {
-			cc = i_read(fd,fp,
-					exp_buffer_end,
-					bufsiz - buf_length,
-					remtime);
+		/*
+		 * always check first if pattern is already in buffer
+		 */
+		if (first_time) {
+			first_time = FALSE;
+			goto after_read;
 		}
 
-		if (cc == 0) return(EXP_EOF);		/* normal EOF */
-		else if (cc == -1) {			/* abnormal EOF */
+		/*
+		 * check for timeout
+		 */
+		if (remtime < 0 || polled) {
+			debuglog("expect: timeout\r\n");
+			exp_match_end = exp_buffer;
+			return_normally(EXP_TIMEOUT);
+		}
+
+		/*
+		 * if timeout == 0, indicate a poll has
+		 * occurred so that next time through loop causes timeout
+		 */
+		if (exp_timeout == 0) {
+			polled = 1;
+		}
+
+		cc = i_read(fd,fp,
+				exp_buffer_end,
+				bufsiz - buf_length,
+				remtime);
+
+		if (cc == 0) {
+			debuglog("expect: eof\r\n");
+			return_normally(EXP_EOF);	/* normal EOF */
+		} else if (cc == -1) {			/* abnormal EOF */
 			/* ptys produce EIO upon EOF - sigh */
 			if (i_read_errno == EIO) {
 				/* convert to EOF indication */
-				return(EXP_EOF);
+				debuglog("expect: eof\r\n");
+				return_normally(EXP_EOF);
 			}
-			sysreturn(i_read_errno);
-		} else if (cc == -2) return(EXP_TIMEOUT);
+			debuglog("expect: error (errno = %d)\r\n",i_read_errno);
+			return_errno(i_read_errno);
+		} else if (cc == -2) {
+			debuglog("expect: timeout\r\n");
+			exp_match_end = exp_buffer;
+			return_normally(EXP_TIMEOUT);
+		}
 
 		old_length = buf_length;
 		buf_length += cc;
@@ -626,7 +900,9 @@ struct exp_case *ecases;
 		/* be used before being set again, there is no need */
 		exp_buffer_end = exp_buffer + buf_length;
 		*exp_buffer_end = '\0';
+                exp_match_end = exp_buffer;
 
+	after_read:
 		debuglog("expect: does {%s} match ",exp_printify(exp_buffer));
 		/* pattern supplied */
 		for (ec=ecases;ec->type != exp_end;ec++) {
@@ -643,6 +919,7 @@ struct exp_case *ecases;
 			} else if (ec->type == exp_exact) {
 				char *p = strstr(exp_buffer,ec->pattern);
 				if (p) {
+					matched = 1;
 					exp_match = p;
 					exp_match_end = p + strlen(ec->pattern);
 				}
@@ -657,31 +934,50 @@ struct exp_case *ecases;
 					}
 				}
 			} else {
+#ifndef HAVE_TCLREGERROR
 				tclRegexpError = 0;
+#else
+				TclRegError((char *)0);
+#endif
 				if (TclRegExec(ec->re,exp_buffer,exp_buffer)) {
 					matched = 1;
 					exp_match = ec->re->startp[0];
 					exp_match_end = ec->re->endp[0];
+#ifndef HAVE_TCLREGERROR
 				} else if (tclRegexpError) {
 			    		fprintf(stderr,"r.e. match (pattern %s) failed: %s",ec->pattern,tclRegexpError);
+#else
+				} else if (TclGetRegError()) {
+			    		fprintf(stderr,"r.e. match (pattern %s) failed: %s",ec->pattern,TclGetRegError());
+#endif
 				}
 			}
 
 			if (matched != -1) {
 				debuglog("yes\nexp_buffer is {%s}\n",
 						exp_printify(exp_buffer));
-				return(ec->value);
+				return_normally(ec->value);
 			} else debuglog("no\n");
 		}
 
-		/* indicate poll failed */
-		if (exp_timeout == 0) return EXP_TIMEOUT;
-
-		if (exp_timeout != -1) {
+		/*
+		 * Update current time and remaining time.
+		 * Don't bother if we are waiting forever or polling.
+		 */
+		if (exp_timeout > 0) {
 			time(&current_time);
 			remtime = end_time - current_time;
 		}
 	}
+ cleanup:
+	f->buffer     = exp_buffer;
+	f->buffer_end = exp_buffer_end;
+	f->match_end  = exp_match_end;
+	if (sys_error) {
+		errno = sys_error;
+		return -1;
+	}
+	return return_val;
 }
 
 int
@@ -702,8 +998,9 @@ struct exp_case *ecases;
 
 /*VARARGS*/
 int
-exp_expectl(va_alist)
-va_dcl
+exp_expectl EXP_VARARGS_DEF(int,arg1)
+/*exp_expectl(va_alist)*/
+/*va_dcl*/
 {
 	va_list args;
 	int fd;
@@ -711,8 +1008,9 @@ va_dcl
 	int i;
 	enum exp_type type;
 
-	va_start(args);
-	fd = va_arg(args,int);
+	fd = EXP_VARARGS_START(int,arg1,args);
+	/* va_start(args);*/
+	/* fd = va_arg(args,int);*/
 	/* first just count the arg sets */
 	for (i=0;;i++) {
 		type = va_arg(args,enum exp_type);
@@ -737,8 +1035,9 @@ va_dcl
 		sysreturn(ENOMEM);
 
 	/* now set up the actual cases */
-	va_start(args);
-	va_arg(args,int);		/*COMPUTED BUT NOT USED*/
+	fd = EXP_VARARGS_START(int,arg1,args);
+	/*va_start(args);*/
+	/*va_arg(args,int);*/		/*COMPUTED BUT NOT USED*/
 	for (ec=ecases;;ec++) {
 		ec->type = va_arg(args,enum exp_type);
 		if (ec->type == exp_end) break;
@@ -764,8 +1063,9 @@ va_dcl
 }
 
 int
-exp_fexpectl(va_alist)
-va_dcl
+exp_fexpectl EXP_VARARGS_DEF(FILE *,arg1)
+/*exp_fexpectl(va_alist)*/
+/*va_dcl*/
 {
 	va_list args;
 	FILE *fp;
@@ -773,8 +1073,9 @@ va_dcl
 	int i;
 	enum exp_type type;
 
-	va_start(args);
-	fp = va_arg(args,FILE *);
+	fp = EXP_VARARGS_START(FILE *,arg1,args);
+	/*va_start(args);*/
+	/*fp = va_arg(args,FILE *);*/
 	/* first just count the arg-pairs */
 	for (i=0;;i++) {
 		type = va_arg(args,enum exp_type);
@@ -798,8 +1099,7 @@ va_dcl
 					malloc((1+i)*sizeof(struct exp_case))))
 		sysreturn(ENOMEM);
 
-	va_start(args);
-	va_arg(args,FILE *);		/*COMPUTED, BUT NOT USED*/
+	EXP_VARARGS_START(FILE *,arg1,args);	/*COMPUTED, BUT NOT USED*/
 	for (ec=ecases;;ec++) {
 		ec->type = va_arg(args,enum exp_type);
 		if (ec->type == exp_end) break;
