@@ -17,20 +17,51 @@ would appreciate credit if this program or parts of it are used.
 #include "expect_cf.h"
 #include <stdio.h>
 #include <sys/ioctl.h>
+
+#ifdef HAVE_STRREDIR_H
+#include <sys/strredir.h>
+# ifdef SRIOCSREDIR
+#  undef TIOCCONS
+# endif
+#endif
+
+#ifdef HAVE_SYS_FCNTL_H
+#include <sys/fcntl.h>
+#endif
+
 #include "tcl.h"
 #include "exp_rename.h"
 #include "exp_prog.h"
 #include "exp_log.h"
 
+static void
+exp_console_manipulation_failed(s)
+char *s;
+{
+	errorlog("expect: spawn: cannot %s console, check permissions of /dev/console\n",s);
+	exit(-1);
+}
+
 void
 exp_console_set()
 {
+#ifdef SRIOCSREDIR
+	int fd;
+
+	if ((fd = open("/dev/console", O_RDONLY)) == -1) {
+		exp_console_manipulation_failed("open");
+	}
+	if (ioctl(fd, SRIOCSREDIR, 0) == -1) {
+		exp_console_manipulation_failed("redirect");
+	}
+	close(fd);
+#endif
+
 #ifdef TIOCCONS
 	int on = 1;
 
 	if (ioctl(0,TIOCCONS,(char *)&on) == -1) {
-		errorlog(stderr, "expect: spawn: cannot open console, check permissions of /dev/console\n");
-		exit(-1);
+		exp_console_manipulation_failed("redirect");
 	}
 #endif /*TIOCCONS*/
 }
