@@ -6,9 +6,6 @@ Design and implementation of this program was paid for by U.S. tax
 dollars.  Therefore it is public domain.  However, the author and NIST
 would appreciate credit if this program or parts of it are used.
 
-$Revision: 1.18 $
-$Date: 1994/04/11 22:30:15 $
-
 */
 
 #include "expect_cf.h"
@@ -17,6 +14,10 @@ $Date: 1994/04/11 22:30:15 $
 /*#include <sys/time.h> seems to not be present on SVR3 systems */
 /* and it's not used anyway as far as I can tell */
 
+#ifdef HAVE_SYS_WAIT_H
+#include <sys/wait.h>
+#endif
+
 #include <ctype.h>
 
 #include "tcl.h"
@@ -32,15 +33,40 @@ $Date: 1994/04/11 22:30:15 $
 #include "tclRegexp.h"
 #include "exp_regexp.h"
 
-/* tests for if we are running this using /dev/tty */
-#define real_tty_output(x) (exp_stdin_is_tty && ((x)==1) || ((x)==exp_dev_tty))
-#define real_tty_input(x) (exp_stdout_is_tty && ((x)==0) || ((x)==exp_dev_tty))
+#ifndef HAVE_TCLREGERROR
+extern char *tclRegexpError;	/* declared in tclInt.h */
+#else
+extern char *TclGetRegError();
+extern void TclRegError();
+#endif
+
+#define INTER_OUT "interact_out"
+
+/*
+ * tests if we are running this using a real tty
+ *
+ * these tests are currently only used to control what gets written to the
+ * logfile.  Note that removal of the test of "..._is_tty" means that stdin
+ * or stdout could be redirected and yet stdout would still be logged.
+ * However, it's not clear why anyone would use log_file when these are
+ * redirected in the first place.  On the other hand, it is reasonable to
+ * run expect as a daemon in which case, stdin/out do not appear to be
+ * ttys, yet it makes sense for them to be logged with log_file as if they
+ * were.
+ */
+#if 0
+#define real_tty_output(x) (exp_stdout_is_tty && (((x)==1) || ((x)==exp_dev_tty)))
+#define real_tty_input(x) (exp_stdin_is_tty && (((x)==0) || ((x)==exp_dev_tty)))
+#endif
+
+#define real_tty_output(x) (((x)==1) || ((x)==exp_dev_tty))
+#define real_tty_input(x)  (((x)==0) || ((x)==exp_dev_tty))
 
 #define new(x)	(x *)ckalloc(sizeof(x))
 
 struct action {
 	char *statement;
-	int tty_reset;		/* true, reset tty mode upon action */
+	int tty_reset;		/* if true, reset tty mode upon action */
 	int iread;		/* if true, reread indirects */
 	int iwrite;		/* if true, write spawn_id element */
 	int timestamp;		/* if true, generate timestamp */
@@ -84,8 +110,8 @@ static struct action *new_action();
 static int inter_eval();
 
 /* in_keymap() accepts user keystrokes and returns one of MATCH,
-CAN_MATCH, or CANT_MATCH.  These describe whether the keystrokes match
-a key sequence, and could or can't if more characters arrive.  The
+CANMATCH, or CANTMATCH.  These describe whether the keystrokes match a
+key sequence, and could or can't if more characters arrive.  The
 function assigns a matching keymap if there is a match or can-match.
 A matching keymap is assigned on can-match so we know whether to echo
 or not.
@@ -100,7 +126,7 @@ The basic idea of how this works is it does a smart sequential search.
 At each position of the input string, we attempt to match each of the
 keymaps.  If at least one matches, the first match is returned.
 
-If there is a can-match and there are more keymaps to try, we continue
+If there is a CANMATCH and there are more keymaps to try, we continue
 trying.  If there are no more keymaps to try, we stop trying and
 return with an indication of the first keymap that can match.
 
@@ -112,13 +138,14 @@ we're ready).  The other is to return can-match.
 */
 
 static int
-in_keymap(string,stringlen,keymap,km_match,match_length,skip)
+in_keymap(string,stringlen,keymap,km_match,match_length,skip,rm_nulls)
 char *string;
 int stringlen;
 struct keymap *keymap;		/* linked list of keymaps */
 struct keymap **km_match;	/* keymap that matches or can match */
 int *match_length;		/* # of chars that matched */
 int *skip;			/* # of chars to skip */
+int rm_nulls;			/* skip nulls if true */
 {
 	struct keymap *km;
 	char *ks;		/* string from a keymap */
@@ -143,7 +170,7 @@ int *skip;			/* # of chars to skip */
 /* skip over nulls - Pascal Meheut, pascal@cnam.cnam.fr 18-May-1993 */
 /*    for (start_search = string;*start_search;start_search++) {*/
     for (start_search = string;start_search<string_end;start_search++) {
-	if (*km_match) break; /* if we've already found a CAN_MATCH */
+	if (*km_match) break; /* if we've already found a CANMATCH */
 			/* don't bother starting search from positions */
 			/* further along the string */
 
@@ -171,14 +198,20 @@ int *skip;			/* # of chars to skip */
 			/* if we ran out of user-supplied characters, and */
 			/* still haven't matched, it might match if the user */
 			/* supplies more characters next time */
+
 			if (s == string_end) {
 				/* skip to next key entry, but remember */
 				/* possibility that this entry might match */
 				if (!*km_match) *km_match = km;
 				break;
 			}
-			/* skip to next key entry, if characters don't match */
-			if ((*s & 0x7f) != *ks) break;
+
+			if ((*s & 0x7f) == *ks) continue;
+			if ((*s == '\0') && rm_nulls) {
+				ks--;
+				continue;
+			}
+			break;
 		}
 	    } else {
 		/* regexp */
@@ -209,8 +242,30 @@ int *skip;			/* # of chars to skip */
 
 	if (*km_match) {
 		/* report a can-match */
+
+		char *p;
+
 		*skip = (start_search-string)-1;
+#if 0
 		*match_length = stringlen - *skip;
+#else
+		/*
+		 * there may be nulls in the string in which case
+		 * the pattern matchers can report CANMATCH when
+		 * the null is hit.  So find the null and compute
+		 * the length of the possible match.
+		 *
+		 * Later, after we squeeze out the nulls, we will
+		 * retry the match, but for now, go along with
+		 * calling it a CANMATCH
+		 */
+		p = start_search;
+		while (*p) {
+			p++;
+		}
+		*match_length = (p - start_search) + 1;
+		/*printf(" match_length = %d\n",*match_length);*/
+#endif
 		return(EXP_CANMATCH);
 	}
 
@@ -321,14 +376,14 @@ int master;
 
 static int
 update_interact_fds(interp,fd_count,fd_to_input,fd_list,input_base,
-			do_indirect,close_count,real_tty_caller)
+			do_indirect,config_count,real_tty_caller)
 Tcl_Interp *interp;
 int *fd_count;
 struct input ***fd_to_input;	/* map from fd's to "struct input"s */
 int **fd_list;
 struct input *input_base;
 int do_indirect;		/* if true do indirects */
-int *close_count;
+int *config_count;
 int *real_tty_caller;
 {
 	struct input *inp;
@@ -338,7 +393,7 @@ int *real_tty_caller;
 
 	int real_tty = FALSE;
 
-	*close_count = exp_close_count;
+	*config_count = exp_configure_count;
 
 	count = 0;
 	for (inp = input_base;inp;inp=inp->next) {
@@ -412,8 +467,6 @@ int *real_tty_caller;
 	return TCL_OK;
 }
 
-static int inter_indirect_key = 0;
-
 /*ARGSUSED*/
 static char *
 inter_updateproc(clientData, interp, name1, name2, flags)
@@ -423,14 +476,14 @@ char *name1;		/* Name of variable. */
 char *name2;		/* Second part of variable name. */
 int flags;		/* Information about what happened. */
 {
-	inter_indirect_key++;
+	exp_configure_count++;
 	return 0;
 }
 			
 #define finish(x)	{ status = x; goto done; }
 
-static char interpreter_cmd[] = "interpreter";
 static char return_cmd[] = "return";
+static char interpreter_cmd[] = "interpreter";
 
 /*ARGSUSED*/
 int
@@ -450,9 +503,11 @@ char **argv;
 	struct input **fd_to_input;	/* map from fd's to "struct input"s */
 	int *fd_list;
 	struct keymap *km;	/* ptr for above while parsing */
- 	extern char *tclRegexpError;	/* declared in tclInt.h */
+/* 	extern char *tclRegexpError;	/* declared in tclInt.h */
 	int master = EXP_SPAWN_ID_BAD;
 	char *master_string = 0;/* string representation of master */
+	int need_to_close_master = FALSE;	/* if an eof is received */
+				/* we use this to defer close until later */
 
 	int next_tty_reset = FALSE;	/* if we've seen a single -reset */
 	int next_iread = FALSE;/* if we've seen a single -iread */
@@ -479,8 +534,6 @@ char **argv;
 
 	char *replace_user_by_process = 0; /* for -u flag */
 
-	int close_count;	/* monitor exp_close_count */
-
 	struct input *input_base;
 #define input_user input_base
 	struct input *input_default;
@@ -499,15 +552,15 @@ char **argv;
 	struct keymap **end_km;
 
 	int key;
-	int indirect_key;
+	int configure_count;	/* monitor reconfigure events */
 
 	if ((argc == 2) && exp_one_arg_braced(argv[1])) {
-		return(exp_eval_with_one_arg(clientData,interp,argc,argv));
+		return(exp_eval_with_one_arg(clientData,interp,argv));
 	} else if ((argc == 3) && streq(argv[1],"-brace")) {
 		char *new_argv[2];
 		new_argv[0] = argv[0];
 		new_argv[1] = argv[2];
-		return(exp_eval_with_one_arg(clientData,interp,argc,new_argv));
+		return(exp_eval_with_one_arg(clientData,interp,new_argv));
 	}
 
 	argv++;
@@ -798,6 +851,13 @@ char **argv;
 				debuglog("-timestamp is deprecated, use exp_timestamp command\r\n");
 				next_timestamp = TRUE;
 				continue;
+			} else if (exp_flageq("nobrace",arg,7)) {
+				/* nobrace does nothing but take up space */
+				/* on the command line which prevents */
+				/* us from re-expanding any command lines */
+				/* of one argument that looks like it should */
+				/* be expanded to multiple arguments. */
+				continue;
 			}
 		}
 
@@ -832,9 +892,18 @@ char **argv;
 		km->null = FALSE;
 		km->re = 0;
 		if (next_re) {
+#ifndef HAVE_TCLREGERROR
+			tclRegexpError = 0;
+#else
+			TclRegError((char *)0);
+#endif
 			if (0 == (km->re = TclRegComp(*argv))) {
 				exp_error(interp,"bad regular expression: %s",
-					tclRegexpError);
+#ifndef HAVE_TCLREGERROR
+								tclRegexpError);
+#else
+								TclGetRegError());
+#endif
 				return(TCL_ERROR);
 			}
 			next_re = FALSE;
@@ -962,7 +1031,7 @@ char **argv;
 	/* "finish()" to terminate this procedure                      */
 	/***************************************************************/
 
-	status = update_interact_fds(interp,&input_count,&fd_to_input,&fd_list,input_base,1,&close_count,&real_tty);
+	status = update_interact_fds(interp,&input_count,&fd_to_input,&fd_list,input_base,1,&configure_count,&real_tty);
 	if (status == TCL_ERROR) finish(TCL_ERROR);
 
 	if (real_tty) {
@@ -980,8 +1049,8 @@ char **argv;
 
 	key = expect_key++;
 
-	/* declare ourselves "in sync" with external view of indirects */
-	indirect_key = inter_indirect_key;
+	/* declare ourselves "in sync" with external view of close/indirect */
+	configure_count = exp_configure_count;
 
 #ifndef SIMPLE_EVENT
 	/* loop waiting (in event handler) for input */
@@ -992,6 +1061,7 @@ char **argv;
 			/* refined by matcher. */
 		int cc;	/* chars count from read() */
 		int m;	/* master */
+		int m_out; /* where master echoes to */
 		struct action *action = 0;
 		time_t previous_time;
 		time_t current_time;
@@ -1029,21 +1099,18 @@ char **argv;
 			/* the count-down code is broken) */
 		}
 
-		/* update the world, if:
-		   indirects have been updated or
-		   any fds have closed
-		 */
-		if (indirect_key != inter_indirect_key
-		    || exp_close_count != close_count) {
+		/* update the world, if necessary */
+		if (configure_count != exp_configure_count) {
 			status = update_interact_fds(interp,&input_count,
 					&fd_to_input,&fd_list,input_base,1,
-					&close_count,&real_tty);
+					&configure_count,&real_tty);
 			if (status) finish(status);
 		}
 
 		rc = exp_get_next_event(interp,fd_list,input_count,&m,timeout,key);
-		if (rc == EXP_TCLERROR)
-			return(TCL_ERROR);
+		if (rc == EXP_TCLERROR) return(TCL_ERROR);
+
+		if (rc == EXP_RECONFIGURE) continue;
 
 		if (rc == EXP_TIMEOUT) {
 			if (timeout_simple) {
@@ -1084,6 +1151,17 @@ char **argv;
 
 		switch (rc) {
 		case EXP_DATA_NEW:
+			if (u->size == u->msize) {
+			    /* In theory, interact could be invoked when this situation */
+			    /* already exists, hence the "probably" in the warning below */
+
+			    debuglog("WARNING: interact buffer is full, probably because your\r\n");
+			    debuglog("patterns have matched all of it but require more chars\r\n");
+			    debuglog("in order to complete the match.\r\n");
+			    debuglog("Dumping first half of buffer in order to continue\r\n");
+			    debuglog("Recommend you enlarge the buffer or fix your patterns.\r\n");
+			    exp_buffer_shuffle(interp,u,0,INTER_OUT,"interact");
+		        }
 			cc = read(m,	u->buffer + u->size,
 					u->msize - u->size);
 			if (cc > 0) {
@@ -1103,25 +1181,38 @@ char **argv;
 
 				/* avoid another function call if possible */
 				if (debugfile || is_debugging) {
-					debuglog("spawn id %d sent {%s}\r\n",m,
+					debuglog("spawn id %d sent <%s>\r\n",m,
 						exp_printify(u->buffer + u->size - cc));
 				}
 				break;
 			}
-			/*FALLTHRU*/
 
+			rc = EXP_EOF;
 			/* Most systems have read() return 0, allowing */
 			/* control to fall thru and into this code.  On some */
 			/* systems (currently HP and new SGI), read() does */
 			/* see eof, and it must be detected earlier.  Then */
 			/* control jumps directly to this EXP_EOF label. */
+
+			/*FALLTHRU*/
 		case EXP_EOF:
 			action = inp->action_eof;
 			attempt_match = FALSE;
 			skip = u->size;
-			rc = EXP_EOF;
 			debuglog("interact: received eof from spawn_id %d\r\n",m);
+			/* actual close is done later so that we have a */
+			/* chance to flush out any remaining characters */
+			need_to_close_master = TRUE;
+
+#if EOF_SO
+			/* should really check for remaining chars and */
+			/* flush them but this will only happen in the */
+			/* unlikely scenario that there are partially */
+			/* matched buffered chars. */
+			/* So for now, indicate no chars to skip. */
+			skip = 0;
 			exp_close(interp,m);
+#endif
 			break;
 		case EXP_DATA_OLD:
 			cc = 0;
@@ -1137,15 +1228,14 @@ char **argv;
 
 		if (attempt_match) {
 			rc = in_keymap(u->buffer,u->size,inp->keymap,
-				&km,&match_length,&skip);
+				&km,&match_length,&skip,u->rm_nulls);
 		} else {
 			attempt_match = TRUE;
 		}
 
 		/* put regexp result in variables */
 		if (km && km->re) {
-#define INTER_OUT "interact_out"
-#define out(var,val)  debuglog("expect: set %s(%s) {%s}\r\n",INTER_OUT,var, \
+#define out(var,val)  debuglog("expect: set %s(%s) \"%s\"\r\n",INTER_OUT,var, \
 						dprintify(val)); \
 		    Tcl_SetVar2(interp,INTER_OUT,var,val,0);
 
@@ -1183,29 +1273,34 @@ char **argv;
 			}
 		}
 
-		/* dispose of chars that should be skipped */
+		/*
+		 * dispose of chars that should be skipped
+		 * i.e., chars that cannot possibly be part of a match.
+		 */
 		
-		/* skip is chars not involved in match */
-		/* print is with chars involved in match */
+		/* "skip" is count of chars not involved in match */
+		/* "print" is count with chars involved in match */
 
 		if (km && km->writethru) {
 			print = skip + match_length;
 		} else print = skip;
 
-		/* figure out if we should echo any chars */
+		/*
+		 * echo chars if appropriate
+		 */
 		if (km && km->echo) {
 			int seen;	/* either printed or echoed */
 
 			/* echo to stdout rather than stdin */
-			if (m == 0) m = 1;
+			m_out = (m == 0)?1:m;
 
 			/* write is unlikely to fail, since we just read */
 			/* from same descriptor */
 			seen = u->printed + u->echoed;
 			if (skip >= seen) {
-				write(m,u->buffer+skip,match_length);
+				write(m_out,u->buffer+skip,match_length);
 			} else if ((match_length + skip - seen) > 0) {
-				write(m,u->buffer+seen,match_length+skip-seen);
+				write(m_out,u->buffer+seen,match_length+skip-seen);
 			}
 			u->echoed = match_length + skip - u->printed;
 		}
@@ -1301,11 +1396,17 @@ char **argv;
 			}
 		}
 
+#if EOF_SO
 		/* as long as buffer is still around, null terminate it */
 		if (rc != EXP_EOF) {
 			u->buffer[u->size] = '\0';
 			u->lower [u->size] = '\0';
 		}
+#else
+		u->buffer[u->size] = '\0';
+		u->lower [u->size] = '\0';
+#endif
+
 		/* now update printed based on total amount skipped */
 
 		u->printed -= skip;
@@ -1322,6 +1423,11 @@ char **argv;
 			if (skip >= oldprinted + u->echoed) u->echoed = 0;
 		}
 
+		if (rc == EXP_EOF) {
+			exp_close(interp,m);
+			need_to_close_master = FALSE;
+		}
+
 		if (action) {
 got_action:
 			change = (action && action->tty_reset);
@@ -1456,7 +1562,7 @@ got_action:
 
 				/* avoid another function call if possible */
 				if (debugfile || is_debugging) {
-					debuglog("spawn id %d sent {%s}\r\n",m,
+					debuglog("spawn id %d sent <%s>\r\n",m,
 						exp_printify(u->buffer + u->size - cc));
 				}
 				break;
@@ -1493,7 +1599,7 @@ got_action:
 		/* put regexp result in variables */
 		if (km && km->re) {
 #define INTER_OUT "interact_out"
-#define out(i,val)  debuglog("expect: set %s(%s) {%s}\r\n",INTER_OUT,i, \
+#define out(i,val)  debuglog("expect: set %s(%s) \"%s\"\r\n",INTER_OUT,i, \
 						dprintify(val)); \
 		    Tcl_SetVar2(interp,INTER_OUT,i,val,0);
 
@@ -1665,7 +1771,6 @@ got_action:
 		}
 
 		if (action) {
-got_action_2:
 			te = inter_eval(interp,action,m);
 			switch (te) {
 			case TCL_BREAK:
@@ -1771,7 +1876,7 @@ got_action_2:
 
 				/* avoid another function call if possible */
 				if (debugfile || is_debugging) {
-					debuglog("spawn id %d sent {%s}\r\n",m,
+					debuglog("spawn id %d sent <%s>\r\n",m,
 						exp_printify(u->buffer + u->size - cc));
 				}
 				break;
@@ -1817,11 +1922,6 @@ got_action_2:
 
 		/* put regexp result in variables */
 		if (km && km->re) {
-#define INTER_OUT "interact_out"
-#define out(i,val)  debuglog("expect: set %s(%s) {%s}\r\n",INTER_OUT,i, \
-						dprintify(val)); \
-		    Tcl_SetVar2(interp,INTER_OUT,i,val,0);
-
 			char name[20], value[20];
 			regexp *re = km->re;
 			char match_char;/* place to hold char temporarily */
@@ -1995,7 +2095,6 @@ got_action_2:
 		}
 
 		if (action) {
-got_action_3:
 			change = (action && action->tty_reset);
 			if (change && tty_changed)
 				exp_tty_set(interp,&tty_old,was_raw,was_echo);
@@ -2035,6 +2134,8 @@ got_action_3:
 	}
 #endif /* SIMPLE_EVENT */
 
+	if (need_to_close_master) exp_close(interp,master);
+
 	if (tty_changed) exp_tty_set(interp,&tty_old,was_raw,was_echo);
 	if (oldargv) ckfree((char *)argv);
 	if (fd_list) ckfree((char *)fd_list);
