@@ -17,6 +17,7 @@ main(argc, argv)
 int argc;
 char *argv[];
 {
+	int rc = 0;
 	Tcl_Interp *interp = Tcl_CreateInterp();
 
 	if (Tcl_Init(interp) == TCL_ERROR) {
@@ -32,12 +33,16 @@ char *argv[];
 	exp_parse_argv(interp,argc,argv);
 
 	/* become interactive if requested or "nothing to do" */
-	if (exp_interactive) (void) exp_interpreter(interp);
-	else if (exp_cmdfile) exp_interpret_cmdfile(interp,exp_cmdfile);
-	else if (exp_cmdfilename) exp_interpret_cmdfilename(interp,exp_cmdfilename);
+	if (exp_interactive)
+		(void) exp_interpreter(interp);
+	else if (exp_cmdfile)
+		rc = exp_interpret_cmdfile(interp,exp_cmdfile);
+	else if (exp_cmdfilename)
+		rc = exp_interpret_cmdfilename(interp,exp_cmdfilename);
 
 	/* assert(exp_cmdlinecmds != 0) */
 
-	exp_exit(interp,0);
+	exp_exit(interp,rc);
+	/*NOTREACHED*/
 }
 
