@@ -20,6 +20,11 @@ implement select incorrectly, so I would recommend using poll.
 #include <stdio.h>
 #include <errno.h>
 #include <poll.h>
+#include <sys/types.h>
+
+#ifdef HAVE_SYS_WAIT_H
+#include <sys/wait.h>
+#endif
 
 #ifdef HAVE_UNISTD_H
 #  include <unistd.h>
@@ -151,13 +156,7 @@ int key;
 		if (rc != TCL_OK) return(exp_tcl2_returnvalue(rc));
 
 		/* anything in the environment could have changed */
-		/* so revalidate fds */
-		for (i=0;i<n;i++) {
-			if (!exp_fs[masters[i]].valid) {
-				exp_error(interp,"spawn id %d has become invalid due to an asynchronous event\n",masters[i]);
-				return TCL_ERROR;
-			}
-		}
+		return EXP_RECONFIGURE;
 	}
 
 	pc = poll(fds,n,msec);
@@ -172,7 +171,7 @@ int key;
 				*master_out = masters[rr];
 				return(EXP_DATA_NEW);
 			} else {
-				exp_debuglog(interp,"poll[%d].revent = %s\r\n",
+				exp_debuglog("poll[%d].revent = %s\r\n",
 					fds[masters[rr]],
 					bad_poll_type(fds[rr].revents));
 				return(EXP_EOF);
@@ -214,6 +213,40 @@ int ready_mask;
 {
 }
 
+/* have no idea if this works, but it seems reasonable (given that it is a */
+/* hack).  Let me know if it works, or you fix it to make it work - DEL */
+/* There is no portable way to do sub-msecond sleeps on such a system, so */
+/* do the next best thing (without a busy loop) and fake it: sleep the right */
+/* amount of time over the long run.  Note that while "subtotal" isn't */
+/* reinitialized, it really doesn't matter for such a gross hack as random */
+/* scheduling pauses will easily introduce occasional one second delays. */
+int	/* returns TCL_XXX */
+exp_dsleep(interp,sec)
+Tcl_Interp *interp;
+double sec;
+{
+	static double subtotal = 0;
+	int msec;			/* milliseconds */
+	struct pollfd pf;	/* some systems need this, even tho unused */
+
+	subtotal += sec;
+	/* if less then 1 msec pause, do nothing but remember it */
+	if (subtotal < .1) return TCL_OK;
+	msec = (int)(subtotal*1000);
+	subtotal -= ((double)msec)/1000.0;
+ restart:
+	if (tcl_AsyncReady) {
+		int rc = Tcl_AsyncInvoke(interp,TCL_OK);
+		if (rc != TCL_OK) return(rc);
+	}
+
+	if (-1 == poll(&pf,(unsigned long)0,msec) && errno == EINTR) {
+		goto restart;
+	}
+	return TCL_OK;
+}
+
+#if 0
 /* have no idea if this works, but it seems reasonable (given that it is a */
 /* hack).  Let me know if it works, or you fix it to make it work - DEL */
 /* There is no portable way to do sub-msecond sleeps on such a system, so */
@@ -246,6 +279,7 @@ long usec;		/* microseconds */
 	}
 	return TCL_OK;
 }
+#endif /* 0 */
 
 /* set things up for later calls to event handler */
 void
