@@ -11,10 +11,15 @@ would appreciate credit if this program or parts of it are used.
 #include "expect_cf.h"
 #include "tcl.h"
 #ifdef NO_STDLIB_H
-#include "compat/stdlib.h"
+#  if TCL_MAJOR_VERSION == 7 && TCL_MINOR_VERSION >= 5
+#    include "../compat/stdlib.h"
+#  else
+#    include "compat/stdlib.h"
+#  endif
 #else
-#include <stdlib.h>		/* for malloc */
-#endif
+#  include <stdlib.h>		/* for malloc */
+#endif /*NO_STDLIB_H*/
+#include <ctype.h>
 
 /* generate printable versions of random ASCII strings.  Primarily used */
 /* by cmdExpect when -d forces it to print strings it is examining. */
@@ -29,8 +34,8 @@ char *s;
 
 	if (s == 0) return("<null>");
 
-	/* worst case is every character takes 3 to printify */
-	need = strlen(s)*3 + 1;
+	/* worst case is every character takes 4 to printify */
+	need = strlen(s)*4 + 1;
 	if (need > destlen) {
 		if (dest) ckfree(dest);
 		dest = ckalloc(need);
@@ -44,13 +49,10 @@ char *s;
 			strcpy(d,"\\n");		d += 2;
 		} else if (*s == '\t') {
 			strcpy(d,"\\t");		d += 2;
-		} else if ((unsigned)*s < 0x20) { /* unsigned strips parity */
-			sprintf(d,"\\C%c",*s + '`');	d += 3;
-		} else if (*s == 0x7f) {
-			/* not syntactically correct, but you get the point */
-			strcpy(d,"\\7f");		d += 3;
-		} else {
+		} else if (isascii(*s) && isprint(*s)) {
 			*d = *s;			d += 1;
+		} else {
+			sprintf(d,"\\x%02x",*s & 0xff);	d += 4;
 		}
 	}
 	*d = '\0';
