@@ -31,9 +31,9 @@
  */
 
 #include "tcl.h"
-#include "exp_conf.h"
-#include "regexp.h"
-#include "exp_global.h"
+#include "expect_cf.h"
+#include "exp_prog.h"
+#include "tclRegexp.h"
 #include "exp_regexp.h"
 
 #define NOTSTATIC	/* was at one time, but Expect needs access */
@@ -774,8 +774,9 @@ char *prog;
 {
 	register char *scan;	/* Current node. */
 	char *next;		/* Next node. */
+#ifndef strchr	/* May be #defined to something else */
 	extern char *strchr();
-
+#endif
 	scan = prog;
 #ifdef DEBUG
 	if (scan != NULL && regnarrate)
@@ -792,7 +793,6 @@ char *prog;
 		case BOL:
 			if (reginput != regbol)
 /*				return(0);*/
-				/* not exactly sure about this */
 				return(EXP_CANTMATCH);
 			break;
 		case EOL:
@@ -1008,8 +1008,9 @@ char *p;
 	register int count = 0;
 	register char *scan;
 	register char *opnd;
+#ifndef strchr	/* May be #defined to something else */
 /*DEL*/	extern char *strchr();
-
+#endif
 	scan = reginput;
 	opnd = OPERAND(p);
 	switch (OP(p)) {
