@@ -11,7 +11,9 @@ would appreciate credit if this program or parts of it are used.
 #include <sys/types.h>
 #include <stdio.h>
 #include <signal.h>
+#if 0				/* Handled elsewhere.  */
 #include <varargs.h>
+#endif
 #include <errno.h>
 #include <ctype.h>	/* for isspace */
 #include <time.h>	/* for time(3) */
@@ -26,7 +28,9 @@ would appreciate credit if this program or parts of it are used.
 #endif
 
 #include "tcl.h"
+#if TCL_MAJOR_VERSION == 7 && TCL_MINOR_VERSION < 4
 extern char *tclRegexpError;	/* declared in tclInt.h */
+#endif
 
 #include "string.h"
 
@@ -75,7 +79,7 @@ struct ecase {	/* case for expect command */
 			/* glob or exact match begins */
 	int transfer;	/* if false, leave matched chars in input stream */
 	int indices;	/* if true, write indices */
-	int iwrite;	/* if true write spawn_id */
+/*	int iwrite;*/	/* if true write spawn_id */
 	int iread;	/* if true, reread indirects */
 	int timestamp;	/* if true, write timestamps */
 #define CASE_UNKNOWN	0
@@ -98,6 +102,8 @@ static
 struct exp_cmd_descriptor {
 	int cmdtype;			/* bg, before, after */
 	int duration;			/* permanent or temporary */
+	int timeout_specified_by_flag;	/* if -timeout flag used */
+	int timeout;			/* timeout period if flag used */
 	struct exp_cases_descriptor ecd;
 	struct exp_i *i_list;
 } exp_cmds[3];
@@ -134,7 +140,7 @@ void exp_background_filehandlers_run_all();
 static char *exp_indirect_update1();	/* 1-part Tcl variable names */
 static char *exp_indirect_update2();	/* 2-part Tcl variable names */
 
-static int	exp_i_read _ANSI_ARGS_((Tcl_Interp *,int,int));
+static int	exp_i_read _ANSI_ARGS_((Tcl_Interp *,int,int,int));
 
 #ifdef SIMPLE_EVENT
 /*ARGSUSED*/
@@ -175,8 +181,6 @@ static RETSIGTYPE
 sigint_handler(n)
 int n;			/* unused, for compatibility with STDC */
 {
-	extern Tcl_Interp *exp_interp;
-
 #ifdef REARM_SIG
 	signal(SIGINT,sigint_handler);/* not nec. for BSD, but doesn't hurt */
 #endif
@@ -206,7 +210,7 @@ int n;			/* unused, for compatibility with STDC */
 #endif
 
 	/* if anywhere else in code, prepare to exit */
-	exp_exit((Tcl_Interp *)0,0);
+	exp_exit(exp_interp,0);
 }
 #endif /*0*/
 
@@ -250,7 +254,7 @@ int free_ilist;		/* if we should free ilist */
 
 	if (free_ilist) {
 		ec->i_list->ecount--;
-		if (ec->i_list->ecount < 1)
+		if (ec->i_list->ecount == 0)
 			exp_free_i(interp,ec->i_list,exp_indirect_update2);
 	}
 
@@ -346,36 +350,41 @@ char *p;
 /* returns TCL_whatever */
 /*ARGSUSED*/
 int
-exp_eval_with_one_arg(clientData,interp,argc,argv)
+exp_eval_with_one_arg(clientData,interp,argv)
 ClientData clientData;
 Tcl_Interp *interp;
-int argc;
 char **argv;
 {
 	char *buf;
 	int rc;
 	char *a;
 
-	/* + 2 is for blank separating cmd and null at end */
-	buf = ckalloc(strlen(argv[0]) + strlen(argv[1]) + 2);
+	/* + 11 is for " -nobrace " and null at end */
+	buf = ckalloc(strlen(argv[0]) + strlen(argv[1]) + 11);
+	/* recreate statement (with -nobrace to prevent recursion) */
+	sprintf(buf,"%s -nobrace %s",argv[0],argv[1]);
 
-	/* replace top-level newlines with blanks */
-	for (a=argv[1];*a;) {
+	/*
+	 * replace top-level newlines with blanks
+	 */
+
+	/* Should only be necessary to run over argv[1] and then sprintf */
+	/* that into the buffer, but the ICEM guys insist that writing */
+	/* back over the original arguments makes their Tcl compiler very */
+	/* unhappy. */
+	for (a=buf;*a;) {
 		extern char *TclWordEnd();
 
 		for (;isspace(*a);a++) {
 			if (*a == '\n') *a = ' ';
 		}
-#if TCL_MINOR_VERSION == 0
+#if TCL_MAJOR_VERSION == 7 && TCL_MINOR_VERSION == 0
 		a = TclWordEnd(a,0)+1;
 #else
 		a = TclWordEnd(a,0,(int *)0)+1;
 #endif
 	}
 
-	/* recreate statement */
-	sprintf(buf,"%s %s",argv[0],argv[1]);
-
 	rc = Tcl_Eval(interp,buf);
 
 	ckfree(buf);
@@ -391,7 +400,7 @@ struct ecase *ec;
 	ec->body = 0;
 	ec->transfer = TRUE;
 	ec->indices = FALSE;
-	ec->iwrite = FALSE;
+/*	ec->iwrite = FALSE;*/
 	ec->iread = FALSE;
 	ec->timestamp = FALSE;
 	ec->re = 0;
@@ -449,6 +458,8 @@ char **argv;
 	argv++;
 	argc--;
 
+	eg->timeout_specified_by_flag = FALSE;
+
 	ecase_clear(&ec);
 
 	/* Allocate an array to store the ecases.  Force array even if 0 */
@@ -484,10 +495,18 @@ char **argv;
 			} else if (exp_flageq("regexp",arg,2)) {
 				i++;
 				ec.use = PAT_RE;
+#ifndef HAVE_TCLREGERROR
 				tclRegexpError = 0;
+#else
+				TclRegError((char *)0);
+#endif
 				if (!(ec.re = TclRegComp(argv[i]))) {
 					exp_error(interp,"bad regular expression: %s",
+#ifndef HAVE_TCLREGERROR
 								tclRegexpError);
+#else
+								TclGetRegError());
+#endif
 					goto error;
 				}
 			} else if (exp_flageq("exact",arg,2)) {
@@ -520,7 +539,7 @@ char **argv;
 				ec.indices = TRUE;
 				continue;
 			} else if (exp_flageq("iwrite",arg,2)) {
-				ec.iwrite = TRUE;
+/*				ec.iwrite = TRUE;*/
 				continue;
 			} else if (exp_flageq("iread",arg,2)) {
 				ec.iread = TRUE;
@@ -528,6 +547,23 @@ char **argv;
 			} else if (exp_flageq("timestamp",arg,2)) {
 				ec.timestamp = TRUE;
 				continue;
+			} else if (exp_flageq("timeout",arg,2)) {
+				i++;
+				if (i>=argc) {
+					exp_error(interp,"-timeout requires following # of seconds");
+					goto error;
+				}
+
+				eg->timeout = atoi(argv[i]);
+				eg->timeout_specified_by_flag = TRUE;
+				continue;
+			} else if (exp_flageq("nobrace",arg,7)) {
+				/* nobrace does nothing but take up space */
+				/* on the command line which prevents */
+				/* us from re-expanding any command lines */
+				/* of one argument that looks like it should */
+				/* be expanded to multiple arguments. */
+				continue;
 			} else {
 				exp_error(interp,"usage: unrecognized flag <%s>",arg);
 				goto error;
@@ -620,7 +656,7 @@ struct eval_out {
 /* string match */
 /* returns EXP_X where X is MATCH, NOMATCH, FULLBUFFER, TCLERRROR */
 static int
-eval_case_string(interp,e,m,o,last_f,last_case)
+eval_case_string(interp,e,m,o,last_f,last_case,suffix)
 Tcl_Interp *interp;
 struct ecase *e;
 int m;
@@ -628,6 +664,7 @@ struct eval_out *o;		/* 'output' - i.e., final case of interest */
 /* next two args are for debugging, when they change, reprint buffer */
 struct exp_f **last_f;
 int *last_case;
+char *suffix;
 {
 	struct exp_f *f = exp_fs + m;
 	char *buffer;
@@ -637,7 +674,8 @@ int *last_case;
 
 	/* if master or case changed, redisplay debug-buffer */
 	if ((f != *last_f) || e->Case != *last_case) {
-		debuglog("\r\nexpect: does {%s} (spawn_id %d) match %s ",
+		debuglog("\r\nexpect%s: does \"%s\" (spawn_id %d) match %s ",
+			 	suffix,
 				dprintify(buffer),f-exp_fs,
 				pattern_style[e->use]);
 		*last_f = f;
@@ -645,8 +683,12 @@ int *last_case;
 	}
 
 	if (e->use == PAT_RE) {
-		debuglog("{%s}? ",dprintify(e->pat));
+		debuglog("\"%s\"? ",dprintify(e->pat));
+#ifndef HAVE_TCLREGERROR
 		tclRegexpError = 0;
+#else
+		TclRegError((char *)0);
+#endif
 		if (buffer && TclRegExec(e->re,buffer,buffer)) {
 			o->e = e;
 			o->match = e->re->endp[0]-buffer;
@@ -656,15 +698,20 @@ int *last_case;
 			return(EXP_MATCH);
 		} else {
 			debuglog(no);
+#ifndef HAVE_TCLREGERROR
 			if (tclRegexpError) {
-			    exp_error(interp,"r.e. match failed: %s",tclRegexpError);
+			    exp_error(interp,"-re failed: %s",tclRegexpError);
+#else
+			if (TclGetRegError()) {
+			    exp_error(interp,"-re failed: %s",TclGetRegError());
+#endif
 			    return(EXP_TCLERROR);
 		        }
 		    }
 	} else if (e->use == PAT_GLOB) {
 		int match; /* # of chars that matched */
 
-	        debuglog("{%s}? ",dprintify(e->pat));
+	        debuglog("\"%s\"? ",dprintify(e->pat));
 		if (buffer && (-1 != (match = Exp_StringMatch(
 				buffer,e->pat,&e->simple_start)))) {
 			o->e = e;
@@ -676,7 +723,7 @@ int *last_case;
 		} else debuglog(no);
 	} else if (e->use == PAT_EXACT) {
 		char *p = strstr(buffer,e->pat);
-	        debuglog("{%s}? ",dprintify(e->pat));
+	        debuglog("\"%s\"? ",dprintify(e->pat));
 		if (p) {
 			e->simple_start = p - buffer;
 			o->e = e;
@@ -717,7 +764,7 @@ int *last_case;
 /* sets o.e if successfully finds a matching pattern, eof, timeout or deflt */
 /* returns original status arg or EXP_TCLERROR */
 static int
-eval_cases(interp,eg,m,o,last_f,last_case,status,masters,mcount)
+eval_cases(interp,eg,m,o,last_f,last_case,status,masters,mcount,suffix)
 Tcl_Interp *interp;
 struct exp_cmd_descriptor *eg;
 int m;
@@ -728,6 +775,7 @@ int *last_case;
 int status;
 int *masters;
 int mcount;
+char *suffix;
 {
 	int i;
 	int em;	/* master of ecase */
@@ -783,14 +831,14 @@ int mcount;
 			if (em == EXP_SPAWN_ID_ANY) {
 				/* test against each spawn_id */
 				for (j=0;j<mcount;j++) {
-					status = eval_case_string(interp,e,masters[j],o,last_f,last_case);
+					status = eval_case_string(interp,e,masters[j],o,last_f,last_case,suffix);
 					if (status != EXP_NOMATCH) return(status);
 				}
 			} else {
 				/* reject things immediately from wrong spawn_id */
 				if (em != m) continue;
 
-				status = eval_case_string(interp,e,m,o,last_f,last_case);
+				status = eval_case_string(interp,e,m,o,last_f,last_case,suffix);
 				if (status != EXP_NOMATCH) return(status);
 			}
 		}
@@ -879,10 +927,12 @@ int direct;
 				exp_free_fd_single(tmp);
 
 				/* if last bg ecase, disarm spawn id */
-				if (ecmd->cmdtype == EXP_CMD_BG &&
-				    exp_fs[m].bg_ecount == 0) {
-					exp_disarm_background_filehandler(m);
-					exp_fs[m].bg_interp = 0;
+				if (ecmd->cmdtype == EXP_CMD_BG) {
+					exp_fs[m].bg_ecount--;
+					if (exp_fs[m].bg_ecount == 0) {
+						exp_disarm_background_filehandler(m);
+						exp_fs[m].bg_interp = 0;
+					}
 				}
 
 				continue;
@@ -952,7 +1002,7 @@ struct ecase *ec;
 {
 	if (!ec->transfer) Tcl_AppendElement(interp,"-notransfer");
 	if (ec->indices) Tcl_AppendElement(interp,"-indices");
-	if (ec->iwrite) Tcl_AppendElement(interp,"-iwrite");
+/*	if (ec->iwrite) Tcl_AppendElement(interp,"-iwrite");*/
 	if (!ec->Case) Tcl_AppendElement(interp,"-nocase");
 
 	if (ec->re) Tcl_AppendElement(interp,"-re");
@@ -1167,12 +1217,12 @@ char **argv;
 	struct exp_cmd_descriptor *ecmd = (struct exp_cmd_descriptor *) clientData;
 
 	if ((argc == 2) && exp_one_arg_braced(argv[1])) {
-		return(exp_eval_with_one_arg(clientData,interp,argc,argv));
+		return(exp_eval_with_one_arg(clientData,interp,argv));
 	} else if ((argc == 3) && streq(argv[1],"-brace")) {
 		char *new_argv[2];
 		new_argv[0] = argv[0];
 		new_argv[1] = argv[2];
-		return(exp_eval_with_one_arg(clientData,interp,argc,new_argv));
+		return(exp_eval_with_one_arg(clientData,interp,new_argv));
 	}
 
 	if (argc > 1 && (argv[1][0] == '-')) {
@@ -1251,10 +1301,13 @@ char **argv;
 				/* because of Tcl's variable tracing */
 				strcpy(interp->result,msg);
 				result = TCL_ERROR;
-				goto cleanup;
+				goto indirect_update_abort;
 			}
 		}
 	}
+	/* empty i_lists have to be removed from global eg.i_list */
+	/* before returning, even if during error */
+ indirect_update_abort:
 
 	/*
 	 * New exp_i's that have 0 ecases indicate fd/vars to be deleted.
@@ -1269,6 +1322,7 @@ char **argv;
 		}
 		exp_i = next;
 	}
+	if (result == TCL_ERROR) goto cleanup;
 
 	/*
 	 * arm all new bg direct fds
@@ -1295,7 +1349,7 @@ char **argv;
 
 		if (ecmd->ecd.count) {
 			/* append to end */
-			ecmd->ecd.cases = (struct ecase **)ckrealloc(ecmd->ecd.cases, count * sizeof(struct ecase *));
+			ecmd->ecd.cases = (struct ecase **)ckrealloc((char *)ecmd->ecd.cases, count * sizeof(struct ecase *));
 			start_index = ecmd->ecd.count;
 		} else {
 			/* append to beginning */
@@ -1318,6 +1372,14 @@ char **argv;
  cleanup:
 	if (result == TCL_ERROR) {
 		/* in event of error, free any unreferenced ecases */
+		/* but first, split up i_list so that exp_i's aren't */
+		/* freed twice */
+
+		for (exp_i=eg.i_list;exp_i;) {
+			struct exp_i *next = exp_i->next;
+			exp_i->next = 0;
+			exp_i = next;
+		}
 		free_ecases(interp,&eg,1);
 	} else {
 		if (eg.ecd.cases) ckfree((char *)eg.ecd.cases);
@@ -1378,18 +1440,28 @@ struct exp_f *f;
 	if (new_msize != f->msize) {
 		if (!f->buffer) {
 			/* allocate buffer space for 1st time */
-			f->lower = ckalloc((unsigned)new_msize+1);
 			f->buffer = ckalloc((unsigned)new_msize+1);
+			f->lower = ckalloc((unsigned)new_msize+1);
 			f->size = 0;
 		} else {
 			/* buffer already exists - resize */
+
+			/* if truncated, forget about some data */
+			if (f->size > new_msize) {
+				/* copy end of buffer down */
+				memmove(f->buffer,f->buffer+(f->size - new_msize),new_msize);
+				memmove(f->lower, f->lower +(f->size - new_msize),new_msize);
+				f->size = new_msize;
+
+				f->key = expect_key++;
+			}
+
 			f->buffer = ckrealloc(f->buffer,new_msize+1);
 			f->lower = ckrealloc(f->lower,new_msize+1);
-			/* if truncated, forget about some data */
-			if (f->size >= f->msize) f->size = f->msize-1;
 		}
 		f->msize = new_msize;
 		f->buffer[f->size] = '\0';
+
 	}
 }
 
@@ -1427,18 +1499,22 @@ int key;
 	struct exp_f *f;
 	int cc;
 	int write_count;
+	int tcl_set_flags;	/* if we have to discard chars, this tells */
+				/* whether to show user locally or globally */
 
 	if (masters == 0) {
 		/* we already know the master, just find out what happened */
 		cc = exp_get_next_event_info(interp,*m,masters_max);
+		tcl_set_flags = TCL_GLOBAL_ONLY;
 	} else {
 		cc = exp_get_next_event(interp,masters,masters_max,m,timeout,key);
+		tcl_set_flags = 0;
 	}
 
 	if (cc == EXP_DATA_NEW) {
 		/* try to read it */
 
-		cc = exp_i_read(interp,*m,timeout);
+		cc = exp_i_read(interp,*m,timeout,tcl_set_flags);
 
 		/* the meaning of 0 from i_read means eof.  Muck with it a */
 		/* little, so that from now on it means "no new data arrived */
@@ -1466,6 +1542,8 @@ int key;
 	} else if (cc == EXP_DATA_OLD) {
 		f = exp_fs + *m;
 		cc = 0;
+	} else if (cc == EXP_RECONFIGURE) {
+		return EXP_RECONFIGURE;
 	}
 
 	if (cc == EXP_ABEOF) {	/* abnormal EOF */
@@ -1473,6 +1551,9 @@ int key;
 		if (i_read_errno == EIO) {
 			/* Sun, Cray, BSD, and others */
 			cc = EXP_EOF;
+		} else if (i_read_errno == EINVAL) {
+			/* Solaris 2.4 occasionally returns this */
+			cc = EXP_EOF;
 		} else {
 			if (i_read_errno == EBADF) {
 				exp_error(interp,"bad spawn_id (process died earlier?)");
@@ -1535,13 +1616,41 @@ static int i_read_cc;
 
 /* when buffer fills, copy second half over first and */
 /* continue, so we can do matches over multiple buffers */
-static void
-buffer_shuffle(f)
+void
+exp_buffer_shuffle(interp,f,save_flags,array_name,caller_name)
+Tcl_Interp *interp;
 struct exp_f *f;
+int save_flags;
+char *array_name;
+char *caller_name;
 {
+	char spawn_id[10];	/* enough for a %d */
+	char match_char;	/* place to hold char temporarily */
+				/* uprooted by a NULL */
+
 	int first_half = f->size/2;
 	int second_half = f->size - first_half;
 
+	/*
+	 * allow user to see data we are discarding
+	 */
+
+	sprintf(spawn_id,"%d",f-exp_fs);
+	debuglog("%s: set %s(spawn_id) \"%s\"\r\n",
+		 caller_name,array_name,dprintify(spawn_id));
+	Tcl_SetVar2(interp,array_name,"spawn_id",spawn_id,save_flags);
+
+	/* temporarily null-terminate buffer in middle */
+	match_char = f->buffer[first_half];
+	f->buffer[first_half] = 0;
+
+	debuglog("%s: set %s(buffer) \"%s\"\r\n",
+		 caller_name,array_name,dprintify(f->buffer));
+	Tcl_SetVar2(interp,array_name,"buffer",f->buffer,save_flags);
+
+	/* remove middle-null-terminator */
+	f->buffer[first_half] = match_char;
+
 	memcpy(f->buffer,f->buffer+first_half,second_half);
 	memcpy(f->lower, f->lower +first_half,second_half);
 	f->size = second_half;
@@ -1561,6 +1670,7 @@ int x;
 	case TCL_BREAK:			return EXP_TCLBRK;
 	case TCL_CONTINUE:		return EXP_TCLCNT;
 	case EXP_CONTINUE:		return EXP_TCLCNTEXP;
+	case EXP_CONTINUE_TIMER:	return EXP_TCLCNTTIMER;
 	case EXP_TCL_RETURN:		return EXP_TCLRETTCL;
 	}
 }
@@ -1576,6 +1686,7 @@ int x;
 	case EXP_TCLBRK:		return TCL_BREAK;
 	case EXP_TCLCNT:		return TCL_CONTINUE;
 	case EXP_TCLCNTEXP:		return EXP_CONTINUE;
+	case EXP_TCLCNTTIMER:		return EXP_CONTINUE_TIMER;
 	case EXP_TCLRETTCL:		return EXP_TCL_RETURN;
 	}
 }
@@ -1586,20 +1697,22 @@ int x;
 /* the read will complete immediately. */
 /*ARGSUSED*/
 static int
-exp_i_read(interp,m,timeout)
+exp_i_read(interp,m,timeout,save_flags)
 Tcl_Interp *interp;
 int m;
 int timeout;
+int save_flags;
 {
 	struct exp_f *f;
 	int cc = EXP_TIMEOUT;
 
 	f = exp_fs + m;
-	if (f->size == f->msize) buffer_shuffle(f);
-
- restart:
+	if (f->size == f->msize) 
+		exp_buffer_shuffle(interp,f,save_flags,EXPECT_OUT,"expect");
 
 #ifdef SIMPLE_EVENT
+ restart:
+
 	alarm_fired = FALSE;
 
 	if (timeout > -1) {
@@ -1681,6 +1794,7 @@ struct exp_fd_list **fd_union;
 
 		/* for each fd in the i_list */
 		for (fdl=p->fd_list;fdl;fdl=fdl->next) {
+			struct exp_fd_list *tmpfdl;
 			struct exp_fd_list *u;
 
 			if (fdl->fd == EXP_SPAWN_ID_ANY) continue;
@@ -1690,9 +1804,9 @@ struct exp_fd_list **fd_union;
 				if (fdl->fd == u->fd) goto found;
 			}
 			/* if not found, link in as head of list */
-			fdl = exp_new_fd(fdl->fd);
-			fdl->next = *fd_union;
-			*fd_union = fdl;
+			tmpfdl = exp_new_fd(fdl->fd);
+			tmpfdl->next = *fd_union;
+			*fd_union = tmpfdl;
 		found:;
 		}
 	}
@@ -1709,10 +1823,13 @@ int cmdtype;
 	case EXP_CMD_BEFORE: return("expect_before");
 	case EXP_CMD_AFTER: return("expect_after");
 	}
+#ifdef LINT
+	return("unknown expect command");
+#endif
 }
 
-static int exp_indirect_key = 0;
-
+/* exp_indirect_update2 is called back via Tcl's trace handler whenever */
+/* an indirect spawn id list is changed */
 /*ARGSUSED*/
 static char *
 exp_indirect_update2(clientData, interp, name1, name2, flags)
@@ -1722,9 +1839,15 @@ char *name1;		/* Name of variable. */
 char *name2;		/* Second part of variable name. */
 int flags;		/* Information about what happened. */
 {
+	char *msg;
+
 	struct exp_i *exp_i = (struct exp_i *)clientData;
-	exp_indirect_key++;
-	return exp_indirect_update1(interp,&exp_cmds[exp_i->cmdtype],exp_i);
+	exp_configure_count++;
+	msg = exp_indirect_update1(interp,&exp_cmds[exp_i->cmdtype],exp_i);
+
+	exp_background_filehandlers_run_all();
+
+	return msg;
 }
 
 static char *
@@ -1804,7 +1927,7 @@ exp_background_filehandlers_run_all()
 
 		/* is bg_interp the best way to check if armed? */
 		if (f->bg_interp && (f->size > 0)) {
-			exp_background_filehandler((ClientData)f,0/*ignored*/);
+			exp_background_filehandler((ClientData)f->fd_ptr,0/*ignored*/);
 		}
 	}
 }
@@ -1831,14 +1954,22 @@ int mask;
 	int last_case;		/* as above but for case */
 
 	/* restore our environment */
-	f = (struct exp_f *)clientData;
-	m = f - exp_fs;
+	m = *(int *)clientData;
+	f = exp_fs + m;
 	interp = f->bg_interp;
 
 	/* temporarily prevent this handler from being invoked again */
 	exp_block_background_filehandler(m);
 
-	cc = expect_read(interp,(int *)0,1,&m,EXP_TIME_INFINITY,0);
+	/* if mask == 0, then we've been called because the patterns changed */
+	/* not because the waiting data has changed, so don't actually do */
+	/* any I/O */
+
+	if (mask == 0) {
+		cc = 0;
+	} else {
+		cc = expect_read(interp,(int *)0,mask,&m,EXP_TIME_INFINITY,0);
+	}
 
 do_more_data:
 	eo.e = 0;		/* no final case yet */
@@ -1862,21 +1993,16 @@ do_more_data:
 		/* normal case, got data */
 		/* new data if cc > 0, same old data if cc == 0 */
 
-#if 0
-		f = exp_fs + m;
-#endif
-
 		/* below here, cc as general status */
 		cc = EXP_NOMATCH;
-
 	}
 
 	cc = eval_cases(interp,&exp_cmds[EXP_CMD_BEFORE],
-					m,&eo,&last_f,&last_case,cc,&m,1);
+			m,&eo,&last_f,&last_case,cc,&m,1,"_background");
 	cc = eval_cases(interp,&exp_cmds[EXP_CMD_BG],
-					m,&eo,&last_f,&last_case,cc,&m,1);
-	cc = eval_cases(interp,&exp_cmds[EXP_CMD_BEFORE],
-					m,&eo,&last_f,&last_case,cc,&m,1);
+			m,&eo,&last_f,&last_case,cc,&m,1,"_background");
+	cc = eval_cases(interp,&exp_cmds[EXP_CMD_AFTER],
+			m,&eo,&last_f,&last_case,cc,&m,1,"_background");
 	if (cc == EXP_TCLERROR) {
 		/* only likely problem here is some internal regexp botch */
 		exp_background_error(interp);
@@ -1888,7 +2014,7 @@ do_more_data:
 		eo.f = exp_fs + m;
 		eo.match = eo.f->size;
 		eo.buffer = eo.f->buffer;
-		debuglog("expect: read eof\r\n");
+		debuglog("expect_background: read eof\r\n");
 		goto matched;
 	}
 	if (!eo.e) {
@@ -1897,22 +2023,23 @@ do_more_data:
 	}
 
  matched:
-#define out(i,val)  debuglog("expect: set %s(%s) {%s}\r\n",EXPECT_OUT,i, \
+#define out(i,val)  debuglog("expect_background: set %s(%s) \"%s\"\r\n",EXPECT_OUT,i, \
 						dprintify(val)); \
-		    Tcl_SetVar2(interp,EXPECT_OUT,i,val,0);
+		    Tcl_SetVar2(interp,EXPECT_OUT,i,val,TCL_GLOBAL_ONLY);
 	{
-		int iwrite = FALSE;	/* write spawn_id? */
+/*		int iwrite = FALSE;*/	/* write spawn_id? */
 		char *body = 0;
 		char *buffer;	/* pointer to normal or lowercased data */
 		struct ecase *e = 0;	/* points to current ecase */
 		int match = -1;		/* characters matched */
 		char match_char;	/* place to hold char temporarily */
 					/* uprooted by a NULL */
+		char *eof_body = 0;
 
 		if (eo.e) {
 			e = eo.e;
 			body = e->body;
-			iwrite = e->iwrite;
+/*			iwrite = e->iwrite;*/
 			if (cc != EXP_TIMEOUT) {
 				f = eo.f;
 				match = eo.match;
@@ -2010,7 +2137,7 @@ do_more_data:
 				sprintf(value,"%d",match-1);
 				out("0,end",value);
 			} else if (e && e->use == PAT_FULLBUFFER) {
-				debuglog("expect: full buffer\r\n");
+				debuglog("expect_background: full buffer\r\n");
 			}
 		}
 
@@ -2019,10 +2146,10 @@ do_more_data:
 		if (eo.f) {
 			char spawn_id[10];	/* enough for a %d */
 
-			if (iwrite) {
+/*			if (iwrite) {*/
 				sprintf(spawn_id,"%d",f-exp_fs);
 				out("spawn_id",spawn_id);
-			}
+/*			}*/
 
 			/* save buf[0..match] */
 			/* temporarily null-terminate string in middle */
@@ -2038,20 +2165,32 @@ do_more_data:
 				f->size -= match;
 				f->printed -= match;
 				if (f->size != 0) {
-				   memcpy(f->buffer,f->buffer+match,f->size);
-				   memcpy(f->lower,f->lower+match,f->size);
+				   memmove(f->buffer,f->buffer+match,f->size);
+				   memmove(f->lower,f->lower+match,f->size);
 				}
 				f->buffer[f->size] = '\0';
 				f->lower[f->size] = '\0';
 			}
 
-			if (cc == EXP_EOF) exp_close(interp,f - exp_fs);
+			if (cc == EXP_EOF) {
+				/* exp_close() deletes all background bodies */
+				/* so save eof body temporarily */
+				if (body) {
+					eof_body = ckalloc(strlen(body)+1);
+					strcpy(eof_body,body);
+					body = eof_body;
+				}
+
+				exp_close(interp,f - exp_fs);
+			}
 
 		}
 
 		if (body) {
-			int result = Tcl_Eval(interp,body);
+			int result = Tcl_GlobalEval(interp,body);
 			if (result != TCL_OK) exp_background_error(interp);
+
+			if (eof_body) ckfree(eof_body);
 		}
 
 
@@ -2111,24 +2250,24 @@ char **argv;
 	int first_time = 1;	/* if not "restarted" */
 
 	int key;		/* identify this expect command instance */
-	int indirect_key;
-
-	int close_count;	/* monitor exp_close_count */
+	int configure_count;	/* monitor exp_configure_count */
 
 	int timeout;		/* seconds */
 	int remtime;		/* remaining time in timeout */
+	int reset_timer;	/* should timer be reset after continue? */
 
 	if ((argc == 2) && exp_one_arg_braced(argv[1])) {
-		return(exp_eval_with_one_arg(clientData,interp,argc,argv));
+		return(exp_eval_with_one_arg(clientData,interp,argv));
 	} else if ((argc == 3) && streq(argv[1],"-brace")) {
 		char *new_argv[2];
 		new_argv[0] = argv[0];
 		new_argv[1] = argv[2];
-		return(exp_eval_with_one_arg(clientData,interp,argc,new_argv));
+		return(exp_eval_with_one_arg(clientData,interp,new_argv));
 	}
 
 	time(&start_time_total);
 	start_time = start_time_total;
+	reset_timer = TRUE;
 
 	/* make arg list for processing cases */
 	/* do it dynamically, since expect can be called recursively */
@@ -2141,9 +2280,6 @@ char **argv;
 		return TCL_ERROR;
 
  restart_with_update:
-	/* get the latest timeout */
-	timeout = get_timeout(interp);
-
 	/* validate all descriptors */
 	/* and flatten fds into array */
 
@@ -2154,9 +2290,8 @@ char **argv;
 		goto cleanup;
 	}
 
-	close_count = exp_close_count;
-	/* declare ourselves "in sync" with external view of indirects */
-	indirect_key = exp_indirect_key;
+	/* declare ourselves "in sync" with external view of close/indirect */
+	configure_count = exp_configure_count;
 
 	/* count and validate fd_list */
 	mcount = 0;
@@ -2179,6 +2314,13 @@ char **argv;
 	if (first_time) first_time = 0;
 	else time(&start_time);
 
+	if (eg.timeout_specified_by_flag) {
+		timeout = eg.timeout;
+	} else {
+		/* get the latest timeout */
+		timeout = get_timeout(interp);
+	}
+
 	key = expect_key++;
 
 	result = TCL_OK;
@@ -2192,8 +2334,13 @@ char **argv;
 
 	/* timeout code is a little tricky, be very careful changing it */
 	if (timeout != EXP_TIME_INFINITY) {
-		time(&current_time);
-		end_time = current_time + timeout;
+		/* if exp_continue -continue_timer, do not update end_time */
+		if (reset_timer) {
+			time(&current_time);
+			end_time = current_time + timeout;
+		} else {
+			reset_timer = TRUE;
+		}
 	}
 
 	/* remtime and current_time updated at bottom of loop */
@@ -2211,6 +2358,9 @@ char **argv;
 			/* do nothing */
 		} else if (cc == EXP_TIMEOUT) {
 			debuglog("expect: timed out\r\n");
+		} else if (cc == EXP_RECONFIGURE) {
+			reset_timer = FALSE;
+			goto restart_with_update;
 		} else if (cc < 0) { /* EXP_TCLERROR or any other weird value*/
 			goto error;
 		} else {
@@ -2226,11 +2376,11 @@ char **argv;
 		}
 
 		cc = eval_cases(interp,&exp_cmds[EXP_CMD_BEFORE],
-			m,&eo,&last_f,&last_case,cc,masters,mcount);
+			m,&eo,&last_f,&last_case,cc,masters,mcount,"");
 		cc = eval_cases(interp,&eg,
-			m,&eo,&last_f,&last_case,cc,masters,mcount);
+			m,&eo,&last_f,&last_case,cc,masters,mcount,"");
 		cc = eval_cases(interp,&exp_cmds[EXP_CMD_AFTER],
-			m,&eo,&last_f,&last_case,cc,masters,mcount);
+			m,&eo,&last_f,&last_case,cc,masters,mcount,"");
 		if (cc == EXP_TCLERROR) goto error;
 		/* special eof code that cannot be done in eval_cases */
 		/* or above, because it would then be executed several times */
@@ -2259,23 +2409,24 @@ char **argv;
 error:
 	result = exp_2tcl_returnvalue(cc);
  done:
-#define out(i,val)  debuglog("expect: set %s(%s) {%s}\r\n",EXPECT_OUT,i, \
+#define out(i,val)  debuglog("expect: set %s(%s) \"%s\"\r\n",EXPECT_OUT,i, \
 						dprintify(val)); \
 		    Tcl_SetVar2(interp,EXPECT_OUT,i,val,0);
 
 	if (result != TCL_ERROR) {
-		int iwrite = FALSE;	/* write spawn_id? */
+/*		int iwrite = FALSE;*/	/* write spawn_id? */
 		char *body = 0;
 		char *buffer;	/* pointer to normal or lowercased data */
 		struct ecase *e = 0;	/* points to current ecase */
 		int match = -1;		/* characters matched */
 		char match_char;	/* place to hold char temporarily */
 					/* uprooted by a NULL */
+		char *eof_body = 0;
 
 		if (eo.e) {
 			e = eo.e;
 			body = e->body;
-			iwrite = e->iwrite;
+/*			iwrite = e->iwrite;*/
 			if (cc != EXP_TIMEOUT) {
 				f = eo.f;
 				match = eo.match;
@@ -2381,10 +2532,10 @@ error:
 		if (eo.f) {
 			char spawn_id[10];	/* enough for a %d */
 
-			if (iwrite) {
+/*			if (iwrite) {*/
 				sprintf(spawn_id,"%d",f-exp_fs);
 				out("spawn_id",spawn_id);
-			}
+/*			}*/
 
 			/* save buf[0..match] */
 			/* temporarily null-terminate string in middle */
@@ -2400,26 +2551,42 @@ error:
 				f->size -= match;
 				f->printed -= match;
 				if (f->size != 0) {
-				   memcpy(f->buffer,f->buffer+match,f->size);
-				   memcpy(f->lower,f->lower+match,f->size);
+				   memmove(f->buffer,f->buffer+match,f->size);
+				   memmove(f->lower,f->lower+match,f->size);
 				}
 				f->buffer[f->size] = '\0';
 				f->lower[f->size] = '\0';
 			}
 
-			if (cc == EXP_EOF) exp_close(interp,f - exp_fs);
+			if (cc == EXP_EOF) {
+				/* exp_close() deletes all background bodies */
+				/* so save eof body temporarily */
+				if (body) {
+					eof_body = ckalloc(strlen(body)+1);
+					strcpy(eof_body,body);
+					body = eof_body;
+				}
+
+				exp_close(interp,f - exp_fs);
+			}
 
 		}
 
 		if (body) {
 			result = Tcl_Eval(interp,body);
+
+			if (eof_body) ckfree(eof_body);
 		}
 	}
 
  cleanup:
+	if (result == EXP_CONTINUE_TIMER) {
+		reset_timer = FALSE;
+		result = EXP_CONTINUE;
+	}
+
 	if ((result == EXP_CONTINUE)
-	     && (indirect_key == exp_indirect_key)
-	     && (close_count == exp_close_count)) {
+	     && (configure_count == exp_configure_count)) {
 		debuglog("expect: continuing expect\r\n");
 		goto restart;
 	}
@@ -2497,7 +2664,9 @@ char **argv;
 {
 	char *format = 0;
 	time_t seconds = -1;
+	int gmt = FALSE;	/* local time by default */
 	struct tm *tm;
+	Tcl_DString dstring;
 
 	argc--; argv++;
 
@@ -2512,6 +2681,9 @@ char **argv;
 			if (!*argv) goto usage_error;
 			seconds = atoi(*argv);
 			argc--; argv++;
+		} else if (streq(*argv,"-gmt")) {
+			gmt = TRUE;
+			argc--; argv++;
 		} else break;
 	}
 
@@ -2521,9 +2693,17 @@ char **argv;
 		time(&seconds);
 	}
 
+	Tcl_DStringInit(&dstring);
+
 	if (format) {
-		tm = localtime(&seconds);
-		exp_strftime(interp->result,TCL_RESULT_SIZE,format,tm);
+		if (gmt) {
+			tm = gmtime(&seconds);
+		} else {
+			tm = localtime(&seconds);
+		}
+/*		exp_strftime(interp->result,TCL_RESULT_SIZE,format,tm);*/
+		exp_strftime(format,tm,&dstring);
+		Tcl_DStringResult(interp,&dstring);
 	} else {
 		sprintf(interp->result,"%ld",seconds);
 	}
