@@ -5,27 +5,19 @@
 #
 #
 # Copyright (c) 1991-1993 The Regents of the University of California.
-# All rights reserved.
+# Copyright (c) 1994 Sun Microsystems, Inc.
 #
-# Permission is hereby granted, without written agreement and without
-# license or royalty fees, to use, copy, modify, and distribute this
-# software and its documentation for any purpose, provided that the
-# above copyright notice and the following two paragraphs appear in
-# all copies of this software.
-#
-# IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
-# DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
-# OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
-# CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
-#
-# THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
-# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
-# AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
-# ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
-# PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
+# See the file "license.terms" for information on usage and redistribution
+# of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 #
 
-set auto_path [info library]
+set auto_path [list [info library]]
+if {[info commands exec] == ""} {
+    # Some machines, such as the Macintosh, do not have exec 
+    set auto_noexec 1
+}
+set errorCode ""
+set errorInfo ""
 
 # unknown:
 # Invoked when a Tcl command is invoked that doesn't exist in the
@@ -33,18 +25,25 @@ set auto_path [info library]
 #
 #	1. See if the autoload facility can locate the command in a
 #	   Tcl script file.  If so, load it and execute it.
-#	2. See if the command exists as an executable UNIX program.
-#	   If so, "exec" the command.
-#	3. If the command was invoked at top-level:
-#	    (a) see if the command requests csh-like history substitution
+#	2. If the command was invoked interactively at top-level:
+#	    (a) see if the command exists as an executable UNIX program.
+#		If so, "exec" the command.
+#	    (b) see if the command requests csh-like history substitution
 #		in one of the common forms !!, !<number>, or ^old^new.  If
 #		so, emulate csh's history substitution.
-#	    (b) see if the command is a unique abbreviation for another
+#	    (c) see if the command is a unique abbreviation for another
 #		command.  If so, invoke the command.
 
 proc unknown args {
-    global auto_noexec auto_noload env unknown_pending tcl_interactive;
+    global auto_noexec auto_noload env unknown_pending tcl_interactive
+    global errorCode errorInfo
 
+    # Save the values of errorCode and errorInfo variables, since they
+    # may get modified if caught errors occur below.  The variables will
+    # be restored just before re-executing the missing command.
+
+    set savedErrorCode $errorCode
+    set savedErrorInfo $errorInfo
     set name [lindex $args 0]
     if ![info exists auto_noload] {
 	#
@@ -67,15 +66,35 @@ proc unknown args {
 	    unset unknown_pending
 	}
 	if $msg {
-	    return [uplevel $args]
+	    set errorCode $savedErrorCode
+	    set errorInfo $savedErrorInfo
+	    set code [catch {uplevel $args} msg]
+	    if {$code ==  1} {
+		#
+		# Strip the last five lines off the error stack (they're
+		# from the "uplevel" command).
+		#
+
+		set new [split $errorInfo \n]
+		set new [join [lrange $new 0 [expr [llength $new] - 6]] \n]
+		return -code error -errorcode $errorCode \
+			-errorinfo $new $msg
+	    } else {
+		return -code $code $msg
+	    }
 	}
     }
-    if {([info level] == 1) && ([info script] == "") && $tcl_interactive} {
+    if {([info level] == 1) && ([info script] == "") \
+	    && [info exists tcl_interactive] && $tcl_interactive} {
 	if ![info exists auto_noexec] {
 	    if [auto_execok $name] {
+		set errorCode $savedErrorCode
+		set errorInfo $savedErrorInfo
 		return [uplevel exec >&@stdout <@stdin $args]
 	    }
 	}
+	set errorCode $savedErrorCode
+	set errorInfo $savedErrorInfo
 	if {$name == "!!"} {
 	    return [uplevel {history redo}]
 	}
@@ -112,7 +131,7 @@ proc auto_load cmd {
 
     if [info exists auto_index($cmd)] {
 	uplevel #0 $auto_index($cmd)
-	return 1
+	return [expr {[info commands $cmd] != ""}]
     }
     if [catch {set path $auto_path}] {
 	if [catch {set path $env(TCLLIBPATH)}] {
@@ -186,6 +205,9 @@ proc auto_execok name {
 	return $auto_execs($name)
     }
     foreach dir [split $env(PATH) :] {
+	if {$dir == ""} {
+	    set dir .
+	}
 	if {[file executable $dir/$name] && ![file isdirectory $dir/$name]} {
 	    set auto_execs($name) 1
 	    return 1
@@ -238,7 +260,7 @@ proc auto_mkindex {dir args} {
 	    while {[gets $f line] >= 0} {
 		if [regexp {^proc[ 	]+([^ 	]*)} $line match procName] {
 		    append index "set [list auto_index($procName)]"
-		    append index " \"source \$dir/$file\"\n"
+		    append index " \"source {\$dir/$file}\"\n"
 		}
 	    }
 	    close $f
@@ -251,8 +273,18 @@ proc auto_mkindex {dir args} {
 	    error $msg $info $code
 	}
     }
-    set f [open tclIndex w]
-    puts $f $index nonewline
-    close $f
-    cd $oldDir
+    set f ""
+    set error [catch {
+	set f [open tclIndex w]
+	puts $f $index nonewline
+	close $f
+	cd $oldDir
+    } msg]
+    if $error {
+	set code $errorCode
+	set info $errorInfo
+	catch {close $f}
+	cd $oldDir
+	error $msg $info $code
+    }
 }
